"""
@file
@brief Helpers to run examples created with function
@see fn export2tf2onnx.
"""
import collections
import inspect
import numpy
from onnx.numpy_helper import from_array
from onnx.helper import (
    make_node, make_graph, make_model, set_model_props, make_tensor)
from onnx import AttributeProto
from ..onnx2py_helper import guess_dtype, guess_proto_dtype
from ..onnx_tools import ensure_topological_order


_make_name_id = 0


def make_tf2onnx_code(opset, name=None, op_type=None, domain='',
                      inputs=None, outputs=None, attributes=None,
                      used=None, context=None, mark_inits=None, indent=8,
                      **unused):
    """
    Converts an ONNX operators into :epkg:`tf2onnx` code.

    :param opset: target opset for the conversion (usually unused)
    :param name: node name
    :param op_type: operator type
    :param domain: domain
    :param inputs: inputs
    :param outputs: outputs
    :param attributes: attributes
    :param used: dictionary `{k: v}`,
        list of nodes taking *k* as input
    :param context: whole context
    :param mark_inits: marks initializer as replaced
    :param indent: number of spaces to add on the second
        and following rows
    :return: code as str
    """
    def simplify(name, kind, force=False):
        value = None
        if (used is not None and name in used and
                len(used[name]) == 1 and context is not None):
            inits = context['initializers_dict']
            if name in inits:
                v = inits[name]
                if v.dtype == numpy.int64 and v.size < 10:
                    value = v
                    if name not in mark_inits:
                        mark_inits[name] = []
                    mark_inits[name].append(v)

        if value is None and force:
            inits = context['initializers_dict']
            if name not in inits:
                raise RuntimeError(  # pragma: no cover
                    "Unable to find init %r in %r value=%r." % (
                        name, list(sorted(inits)), value))
            value = inits[name]
        if kind == 'list':
            if value is None:
                return name
            if len(value.shape) == 0:
                return str(value)
            return str(list(value))
        if kind == 'list_var':
            if value is None:
                return f"varx[{name!r}]"
            if len(value.shape) == 0:
                return str(value)
            return str(list(value))
        raise NotImplementedError(
            f"Unknown scenario to simplify ({kind!r}).")

    rows = []
    if op_type == 'Unsqueeze':
        if len(inputs) == 2:
            rows.append(
                "node = GraphBuilder(ctx).make_unsqueeze("
                "{'data': varx[%r], 'axes': %s}, return_node=True)"
                "" % (inputs[0], simplify(inputs[1], 'list_var')))
        else:
            raise NotImplementedError(  # pragma: no cover
                f"Unable to create code for operator {op_type!r} (opset <= 12).")
    elif op_type == 'Squeeze':
        if len(inputs) == 1:
            rows.append(
                "node = GraphBuilder(ctx).make_squeeze("
                "{'data': varx[%r]}, return_node=True)"
                "" % (inputs[0], ))
        elif len(inputs) == 2:
            rows.append(
                "node = GraphBuilder(ctx).make_squeeze("
                "{'data': varx[%r], 'axes': %s}, return_node=True)"
                "" % (inputs[0], simplify(inputs[1], 'list_var')))
        else:
            raise NotImplementedError(  # pragma: no cover
                f"Unable to create code for operator {op_type!r} (opset <= 12).")
    elif op_type == 'Slice':
        atts = dict(zip(['starts', 'ends', 'axes', 'steps'],
                        inputs[1:]))
        text = ", ".join(f"'{k}': {simplify(v, 'list_var')}"
                         for k, v in atts.items())
        if len(inputs) in (3, 4, 5):
            rows.append(
                "node = GraphBuilder(ctx).make_slice("
                "{'data': varx[%r], %s}, return_node=True)"
                "" % (inputs[0], text))
        else:
            raise NotImplementedError(  # pragma: no cover
                f"Unable to create code for operator {op_type!r} (opset <= 12).")
    else:
        if len(attributes) > 0:
            attributes_str = ", ".join(f"{k}={v}" for k, v in attributes)
            attr = f", attr=dict({attributes_str})"
        else:
            attr = ""
        rows.append(
            f"inputs = [{', '.join('varx[%r]' % n for n in inputs)}]")
        sdomain = '' if domain == '' else (f"domain={domain!r}, ")
        rows.append(
            "node = ctx.make_node(%r, inputs=inputs%s, %s"
            "name=make_name(%r))" % (
                op_type, attr, sdomain, name))
    for i, n in enumerate(outputs):
        rows.append("varx[%r] = node.output[%d]" % (n, i))
    if indent > 0:
        sind = " " * indent
        for i in range(1, len(rows)):
            rows[i] = sind + rows[i]
    return "\n".join(rows)


def make_name(name):
    "Creates a unique name."
    global _make_name_id  # pylint: disable=W0603
    name = "%s_%d" % (name, _make_name_id)
    _make_name_id += 1
    return name


def get_max_value(np_dtype):
    "Returns the maximum value for a specific type."
    return numpy.iinfo(np_dtype).max


def make_sure(cond, msg, *args):
    "Raises an exception if cond is not verified."
    if not cond:
        raise RuntimeError(msg % tuple(args))  # pragma: no cover


def map_onnx_to_numpy_type(onnx_dtype):
    "Converts ONNX type into numpy type."
    if onnx_dtype is None:
        return numpy.float32
    return guess_dtype(onnx_dtype)


class tf_op:
    """
    Decorator to register any new converter.
    :param name: type of the operator to rewrite
    :param domain: domain
    """
    _OPSETS = collections.OrderedDict()

    def __init__(self, name, domain='', **kwargs):
        if not isinstance(name, list):
            name = [name]
        self.names = name
        self.domain = domain
        self.kwargs = kwargs

    def __call__(self, func):
        for ke, va in inspect.getmembers(func, inspect.ismethod):
            if ke.startswith("version_"):
                version = int(ke.replace("version_", ""))
                self._register_handler(
                    va, version, self.names, self.domain, self.kwargs)
        return func

    def _register_handler(self, func, version, names, domain, kwargs):
        opset = tf_op._OPSETS.get(domain)
        if not opset:
            opset = []
            tf_op._OPSETS[domain] = opset
        while version >= len(opset):
            opset.append({})
        opset_dict = opset[version]
        for name in names:
            opset_dict[name] = (func, kwargs)


class Tf2OnnxConvert:
    """
    Applies the converter on an ONNX graph.

    :param onnx_model: ONNX graph
    :param tf_op: class which register
    :param verbose: verbosity
    :param target_opset: targetted opsets
    """

    def __init__(self, onnx_model, _tf_op=None, verbose=None,
                 target_opset=None, max_iter=5):
        self._onnx_model = onnx_model
        self._tf_op = _tf_op or tf_op
        self.verbose = verbose
        self.max_iter = max_iter
        if isinstance(target_opset, int):
            self.target_opsets = {'': target_opset}
        elif isinstance(target_opset, dict):
            self.target_opsets = target_opset
        elif target_opset is None:
            opsets = {}
            for oimp in onnx_model.opset_import:
                if oimp.domain == '':
                    opsets[oimp.domain] = oimp.version
                    opset = oimp.version
                else:
                    opsets[oimp.domain] = opset
            self.target_opsets = opsets
        else:
            raise ValueError(  # pragma: no cover
                f"Unexepected value for target_opset={target_opset!r}.")
        self._names = {}
        for node in onnx_model.graph.node:
            self._names[node.name] = node
        for init in onnx_model.graph.initializer:
            self._names[init.name] = init
        # _forbidden_new_names contains current names and deleted names.
        self._forbidden_new_names = set(self._names)
        if '' in self.target_opsets:
            self.opset = self.target_opsets['']
        if not hasattr(self, 'opset'):
            raise RuntimeError(  # pragma: no cover
                f"Attribute opset is missing, target_opset={target_opset!r}.")

    def get_node_by_name(self, name):
        """
        Retrieves a node by its name.

        :param name: node name
        :return: node name
        """
        if name not in self._names:
            raise RuntimeError(  # pragma: no cover
                "Unable to find node name %r among %r." % (
                    name, ", ".join(sorted(self._names))))
        return self._names[name]

    def _add_node_name(self, obj):
        """
        Registers an object in in the graph by its name.
        :param name: node or initializer
        """
        if obj.name in self._forbidden_new_names:
            raise RuntimeError(  # pragma: no cover
                f"Name {obj.name!r} is already registered.")
        self._names[obj.name] = obj
        self._forbidden_new_names.add(obj.name)

    def make_node(self, op_type, inputs, attr=None, outputs=None,
                  name=None, domain='', output_count=1,
                  shapes=None, dtypes=None):
        """
        Adds a node to the list of nodes.

        :param op_type: operator type
        :param inputs: list of strings
        :param attr: dictionary of attributes
        :param outputs: None or list of strings
        :param output_count: used if outputs is None to guess
            the number of outputs of this node
        :param name: name of the node
        :param domain: domain
        :param shapes: unused
        :param dtypes: unused
        :return: created node
        """
        if self.verbose:
            print(  # pragma: no cover
                f"[Tf2OnnxConvert.make_node] op_type={op_type!r} inputs={inputs!r}")

        if attr is None:
            attr = {}
        if name is None:
            name = make_name(op_type)
        if name in self._names:
            raise RuntimeError(  # pragma: no cover
                "Node name %r already exists in %r." % (
                    name, ", ".join(sorted(self._names))))

        if outputs is None:
            outputs = [(name + ":" + str(i)) for i in range(output_count)]

        output_count = len(outputs)
        raw_attr = {}
        onnx_attrs = []
        for a, v in attr.items():
            if isinstance(v, AttributeProto):
                onnx_attrs.append(v)
            else:
                raw_attr[a] = v

        onnx_node = make_node(
            op_type, inputs, outputs, name=name, domain=domain, **raw_attr)

        self._add_node_name(onnx_node)
        return onnx_node

    def make_const(self, name, np_val, skip_conversion=False, raw=True):
        """
        Make a new constants in the graph.
        :param name: const node name, must be unique.
        :param np_val: value of type numpy ndarray.
        :param skip_conversion:
            bool, indicate whether this created node would be mapped
            during conversion
        :param raw: whether to store data at field of raw_data or the
            specific field according to its dtype
        :return: create initializer
        """
        if name in self._names:
            raise RuntimeError(  # pragma: no cover
                "Initializer name %r already exists in %r." % (
                    name, ", ".join(sorted(self._names))))
        np_val_flat = np_val.flatten()
        is_bytes = (np_val.dtype == numpy.object and len(np_val_flat) > 0 and
                    isinstance(np_val_flat[0], bytes))
        if raw and not is_bytes:
            onnx_tensor = from_array(np_val, name)
        else:
            onnx_tensor = make_tensor(
                name, guess_proto_dtype(np_val.dtype),
                np_val.shape, np_val_flat, raw=False)

        self._add_node_name(onnx_tensor)
        return onnx_tensor

    def get_dtype(self, input_name):
        """
        Returns the type of one node or None if unknown.
        :param input_name: result name
        :return: numpy dtype
        """
        inputs = self._onnx_model.graph.input
        names = [_.name for _ in inputs]
        if input_name not in names:
            return None  # pragma: no cover
        ind = names.index(input_name)
        return inputs[ind].type.tensor_type.elem_type

    def replace_all_inputs(self, old_name, new_name):
        """
        Every taking *old_name* as inputs will take *new_name* instead.
        Looks in the output as well but in that case, it creates an identity
        node to avoid changing an output name.
        :param old_name: name to replace
        :param new_name: new name
        :return: list of impacted nodes
        """
        if self.verbose:
            print(  # pragma: no cover
                "[Tf2OnnxConvert.replace_all_inputs] replace %r by %r" % (
                    old_name, new_name))
        res = []
        for node in self._names.values():
            if not hasattr(node, 'input'):
                continue
            if old_name not in node.input:
                continue
            new_inputs = [new_name if i == old_name else i
                          for i in node.input]
            node.input[:] = new_inputs[:]
            res.append(node)
            if self.verbose:
                print(  # pragma: no cover
                    "[Tf2OnnxConvert.replace_all_inputs] replace %r by %r in node %r" % (
                        old_name, new_name, node.name))
        for o in self._onnx_model.graph.output:
            if o.name != old_name:
                continue
            n = self.make_node("Identity", [new_name], outputs=[old_name],
                               name=make_name("IdOutputReplaced"))
            res.append(n)
            if self.verbose:
                print(  # pragma: no cover
                    "[Tf2OnnxConvert.replace_all_inputs] add id node from %r to %r "
                    "with node %r." % (
                        old_name, new_name, n.name))  # pylint: disable=E1101
        if self.verbose:
            print(  # pragma: no cover
                "[Tf2OnnxConvert.replace_all_inputs] end")
        return res

    def remove_node(self, name):
        """
        Removes a node name from the list.
        """
        if name not in self._names:
            raise RuntimeError(  # pragma: no cover
                f"Unable to delete name {name!r} because it does not exists.")
        del self._names[name]
        if self.verbose:
            print(  # pragma: no cover
                f"[Tf2OnnxConvert.remove_node] delete name {name!r}")

    def get_shape(self, input_name):
        """
        Returns the type of one node or None if unknown.
        :param input_name: result name
        :return: numpy dtype
        """
        inputs = self._onnx_model.graph.input
        names = [_.name for _ in inputs]
        if input_name not in names:
            return None  # pragma: no cover
        ind = names.index(input_name)
        dims = inputs[ind].type.tensor_type.shape.dim
        return tuple(dims)

    def run(self):
        """
        Calls the registered converters on the graph
        held by this instance. Returns the new onnx graph.

        :return: ONNX graph
        """
        if len(self._tf_op._OPSETS) == 0:
            raise RuntimeError(  # pragma: no cover
                "No converter was registered.")
        if self.verbose:
            print("[Tf2OnnxConvert.run]")  # pragma: no cover

        done = {}
        modif = 1
        turn = 0
        while modif > 0 and turn < self.max_iter:
            modif = 0
            turn += 1
            # The converter may alter the current list of nodes, we freeze it.
            current_values = list(self._names.values())
            for node in current_values:
                if not hasattr(node, 'domain'):
                    # initializer
                    continue
                if done.get(node.name, False):
                    continue
                domain = node.domain
                if domain not in self._tf_op._OPSETS:
                    continue

                # look for a converter
                rews = self._tf_op._OPSETS[domain]
                target = min(self.target_opsets[domain], len(rews))
                conv = None
                for i in range(len(rews) - 1, -1, -1):
                    if node.op_type in rews[i]:
                        conv = rews[i][node.op_type]
                        break
                if conv is None:
                    continue

                # applies the converter
                if self.verbose:
                    print(  # pragma: no cover
                        "[Tf2OnnxConvert.run] convert node type=%r opset=%r name=%r"
                        "" % (node.op_type, target, node.name))
                fct, kwargs = conv
                fct(self, node, target_opset=target, **kwargs)
                modif += 1

        if turn >= self.max_iter:
            raise RuntimeError(  # pragma: no cover
                "Too many iterations and no stable ONNX was reached, "
                "iter=%d\n%s" % (turn, str(self.make_model())))
        return self.make_model()

    def make_model(self):
        """
        Produces the new ONNX graph with the updated sets of nodes.
        """
        inputs = self._onnx_model.graph.input
        outputs = self._onnx_model.graph.output
        inits = [init[1] for init in sorted(self._names.items())
                 if not hasattr(init[1], 'domain')]
        nodes = [node[1] for node in sorted(self._names.items())
                 if hasattr(node[1], 'domain')]
        nodes = ensure_topological_order(inputs, inits, nodes)

        if self.verbose:
            print(  # pragma: no cover
                "[Tf2OnnxConvert.make_node] %d nodes %d inputs %d "
                "outputs %d initializers"
                "" % (len(nodes), len(inputs), len(outputs), len(inits)))
        graph = make_graph(nodes, self._onnx_model.graph.name,
                           inputs, outputs, inits)
        onnx_model = make_model(graph, functions=self._onnx_model.functions)
        onnx_model.ir_version = self._onnx_model.ir_version
        onnx_model.producer_name = self._onnx_model.producer_name + "-mlprodict"
        onnx_model.producer_version = self._onnx_model.producer_version
        onnx_model.domain = self._onnx_model.domain
        onnx_model.model_version = self._onnx_model.model_version
        onnx_model.doc_string = self._onnx_model.doc_string
        metadata = {p.key: p.value for p in self._onnx_model.metadata_props}
        set_model_props(onnx_model, metadata)

        # opsets
        del onnx_model.opset_import[:]  # pylint: disable=E1101
        for dom, value in self.target_opsets.items():
            op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
            op_set.domain = dom
            op_set.version = value
        return onnx_model


class GraphBuilder:
    """
    Helpers to build graph.
    :param graph!
    """

    def __init__(self, graph):
        self._g = graph

    @property
    def graph(self):
        "Returns the graph."
        return self._g

    def make_slice(self, kwargs, name=None, shapes=None, dtypes=None,
                   return_node=False):
        """
        slice changes its schema at opset 10: it treats some
        attributes as dynamic input so this function has to process
        inputs according to graph's opset version
        to get "inputs" and "attr" to feed "make_node"
        kwargs: key could be `["data", "starts", "ends",
        "axes", "steps", "outputs"]`.
        """
        outputs = kwargs.pop("outputs", None)

        if self.graph.opset < 10:
            # "data" is string
            # "starts", "ends" and "axes" are attributes,
            # and "axes" is optional.
            data = kwargs.pop("data")  # pragma: no cover
            starts = self._convert_to_attribute(  # pragma: no cover
                kwargs.pop("starts"))
            ends = self._convert_to_attribute(  # pragma: no cover
                kwargs.pop("ends"))
            axes = self._convert_to_attribute(  # pragma: no cover
                kwargs.pop("axes", None), is_optional=True)
            attr = {"starts": starts, "ends": ends,
                    "axes": axes}  # pragma: no cover
            inputs = [data]  # pragma: no cover
        else:
            # slice-10 has 3 required inputs "data", "starts", "ends"l
            # and 2 optional inputs "axes", "steps"
            # input sequence should be "data", "starts", "ends",
            # "axes", "steps"
            attr = {}
            data = kwargs.pop("data")
            starts = self._convert_to_input(
                kwargs.pop("starts"), "const_starts", dtype=numpy.int64)
            ends = self._convert_to_input(
                kwargs.pop("ends"), "const_ends", dtype=numpy.int64)
            axes = self._convert_to_input(
                kwargs.pop("axes", None), "const_axes",
                is_optional=True, dtype=numpy.int64)
            steps = self._convert_to_input(
                kwargs.pop("steps", None), "const_steps",
                is_optional=True, dtype=numpy.int64)
            inputs = [data, starts, ends, axes, steps]

        # pro-process inputs and attr
        make_sure(not kwargs, "kwargs contains un-used key")

        new_attr = {}
        for key, val in attr.items():
            if val is not None:
                new_attr[key] = val
        attr = new_attr

        for ind, val in enumerate(inputs):
            if val is None:
                inputs[ind] = ""  # empty string means no connection in ONNX
        # remove tailing ""
        while inputs[-1] == "":
            inputs = inputs[:-1]

        if self.graph.opset >= 10:
            dtype = self.graph.get_dtype(inputs[1])
            for input_data in inputs[1:]:
                if input_data != "":
                    make_sure(dtype == self.graph.get_dtype(
                        input_data), "dtype should be same")

        node = self.graph.make_node(op_type="Slice", inputs=inputs, attr=attr,
                                    name=name, outputs=outputs, shapes=shapes,
                                    dtypes=dtypes)
        if return_node:
            return node
        raise NotImplementedError(  # pragma: no cover
            "return_node must be True")

    def make_squeeze(self, kwargs, name=None, shapes=None, dtypes=None,
                     return_node=False, op_name_scope=None):
        """
        Squeeze changes its schema at opset 13: it treats axes as a dynamic input
        kwargs: key could be ["data", "axes"].
        """
        outputs = kwargs.pop("outputs", None)

        if self.graph.opset < 13:
            data = kwargs.pop("data")
            axes = self._convert_to_attribute(
                kwargs.pop("axes", None), is_optional=True)
            attr = {"axes": axes}
            inputs = [data]
        else:
            data = kwargs.pop("data")
            axes = self._convert_to_input(
                kwargs.pop("axes", None), "const_axes",
                is_optional=True, dtype=numpy.int64)
            attr = {}
            inputs = [data, axes]

        make_sure(not kwargs, "kwargs contains un-used key")

        new_attr = {}
        for key, val in attr.items():
            if val is not None:
                new_attr[key] = val
        attr = new_attr

        for ind, val in enumerate(inputs):
            if val is None:
                inputs[ind] = ""  # empty string means no connection in ONNX
        # remove tailing ""
        while inputs[-1] == "":
            inputs = inputs[:-1]

        node = self.graph.make_node(
            op_type="Squeeze", inputs=inputs, attr=attr, name=name,
            outputs=outputs)
        if return_node:
            return node
        raise NotImplementedError(  # pragma: no cover
            "return_node must be True")

    def make_unsqueeze(self, kwargs, name=None, shapes=None, dtypes=None,
                       return_node=False, op_name_scope=None):
        """
        Unsqueeze changes its schema at opset 13: it treats axes as a dynamic input
        kwargs: key could be ["data", "axes"].
        """
        outputs = kwargs.pop("outputs", None)

        if self.graph.opset < 13:
            data = kwargs.pop("data")  # pragma: no cover
            axes = self._convert_to_attribute(  # pragma: no cover
                kwargs.pop("axes", None), is_optional=True)
            attr = {"axes": axes}  # pragma: no cover
            inputs = [data]  # pragma: no cover
        else:
            data = kwargs.pop("data")
            axes = self._convert_to_input(
                kwargs.pop("axes", None), "const_axes",
                is_optional=True, dtype=numpy.int64)
            attr = {}
            inputs = [data, axes]

        make_sure(not kwargs, "kwargs contains un-used key")

        new_attr = {}
        for key, val in attr.items():
            if val is not None:
                new_attr[key] = val
        attr = new_attr

        for ind, val in enumerate(inputs):
            if val is None:
                inputs[ind] = ""  # empty string means no connection in ONNX
        # remove tailing ""
        while inputs[-1] == "":
            inputs = inputs[:-1]

        node = self.graph.make_node(
            op_type="Unsqueeze", inputs=inputs, attr=attr, name=name,
            outputs=outputs)
        if return_node:
            return node
        raise NotImplementedError(  # pragma: no cover
            "return_node must be True")

    def _convert_to_input(self, tensor, const_name, is_optional=False, dtype=None):
        """in ONNX, input shold come from node, so it must be a string"""
        if is_optional and tensor is None:
            return None

        make_sure(tensor is not None,
                  "input is required so it couldn't be None")

        res = tensor
        if isinstance(tensor, list):
            res = self.graph.make_const(
                make_name(const_name), numpy.array(tensor, dtype)).name
        return res

    def _convert_to_attribute(self, tensor, is_optional=False):
        if is_optional and tensor is None:
            return None

        make_sure(tensor is not None,
                  "input is required so it couldn't be None")

        res = tensor
        if isinstance(tensor, str):
            const_node = self.graph.get_node_by_output(tensor)
            res = const_node.get_tensor_value(as_list=True)

        make_sure(isinstance(res, list),
                  "input is an attr, so a list is needed")

        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Hardmax(OpRunUnaryNum):

    atts = {'axis': -1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Hardmax.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        x_argmax = numpy.argmax(x, axis=self.axis)
        y = numpy.zeros_like(x)
        numpy.put_along_axis(y, numpy.expand_dims(x_argmax, axis=self.axis),
                             1, axis=self.axis)
        return (y, )

    def to_python(self, inputs):
        return ("import numpy",
                "\n".join([
                    "{0}_argmax = numpy.argmax({0}, axis=axis)".format(
                        inputs[0]),
                    "{0}y = numpy.zeros_like({0})".format(inputs[0]),
                    f"numpy.put_along_axis({inputs[0]}y,",
                    "    numpy.expand_dims(",
                    f"       {inputs[0]}_argmax, axis=axis),",
                    "    1, axis=axis)",
                    f"return {inputs[0]}y"]))



"""
@file
@brief Automates the generation of operators for the
documentation for the Xop API.

::

    def setup(app):
        app.connect('builder-inited', generate_op_doc)

.. versionadded:: 0.9
"""
from .xop_auto import onnx_documentation_folder


def _generate_op_doc(app):
    from sphinx.util import logging
    logger = logging.getLogger(__name__)
    folder = app.config.onnx_doc_folder
    onnx_documentation_folder(folder, fLOG=logger.info)


def setup(app):
    """
    Sphinx extension `mlprodict.npy.xop_sphinx` displays documentation
    on ONN Operators.
    """
    import sphinx
    app.add_config_value('onnx_doc_folder', 'onnx_doc_folder', 'env')
    app.connect('builder-inited', _generate_op_doc)
    return {'version': sphinx.__display_version__, 'parallel_read_safe': True}



# -*- encoding: utf-8 -*-
"""
@file
@brief Registers new converters.
"""
import copy
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from skl2onnx import (
    update_registered_converter,
    update_registered_parser)
from skl2onnx.common.data_types import guess_tensor_type
from skl2onnx.common._apply_operation import apply_identity


class CustomScorerTransform(BaseEstimator, TransformerMixin):
    """
    Wraps a scoring function into a transformer. Function @see fn
    register_scorers must be called to register the converter
    associated to this transform. It takes two inputs, expected values
    and predicted values and returns a score for each observation.
    """

    def __init__(self, name, fct, kwargs):
        """
        @param      name        function name
        @param      fct         python function
        @param      kwargs      parameters function
        """
        BaseEstimator.__init__(self)
        TransformerMixin.__init__(self)
        self.name_fct = name
        self._fct = fct
        self.kwargs = kwargs

    def __repr__(self):  # pylint: disable=W0222
        return "{}('{}', {}, {})".format(
            self.__class__.__name__, self.name_fct,
            self._fct.__name__, self.kwargs)


def custom_scorer_transform_parser(scope, model, inputs, custom_parsers=None):
    """
    This function updates the inputs and the outputs for
    a @see cl CustomScorerTransform.

    :param scope: Scope object
    :param model: A scikit-learn object (e.g., *OneHotEncoder*
        or *LogisticRegression*)
    :param inputs: A list of variables
    :param custom_parsers: parsers determines which outputs is expected
        for which particular task, default parsers are defined for
        classifiers, regressors, pipeline but they can be rewritten,
        *custom_parsers* is a dictionary
        ``{ type: fct_parser(scope, model, inputs, custom_parsers=None) }``
    :return: A list of output variables which will be passed to next
        stage
    """
    if custom_parsers is not None:  # pragma: no cover
        raise NotImplementedError(
            "Case custom_parsers not empty is not implemented yet.")
    if isinstance(model, str):
        raise RuntimeError(  # pragma: no cover
            f"Parameter model must be an object not a string '{model}'.")
    if len(inputs) != 2:
        raise RuntimeError(  # pragma: no cover
            f"Two inputs expected not {len(inputs)}.")
    alias = 'Mlprodict' + model.__class__.__name__
    this_operator = scope.declare_local_operator(alias, model)
    this_operator.inputs = inputs

    scores = scope.declare_local_variable(
        'scores', guess_tensor_type(inputs[0].type))
    this_operator.outputs.append(scores)
    return this_operator.outputs


def custom_scorer_transform_shape_calculator(operator):
    """
    Computes the output shapes for a @see cl CustomScorerTransform.
    """
    if len(operator.inputs) != 2:
        raise RuntimeError("Two inputs expected.")  # pragma: no cover
    if len(operator.outputs) != 1:
        raise RuntimeError("One output expected.")  # pragma: no cover

    N = operator.inputs[0].type.shape[0]
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = [N, 1]


def custom_scorer_transform_converter(scope, operator, container):
    """
    Selects the appropriate converter for a @see cl CustomScorerTransform.
    """
    op = operator.raw_operator
    name = op.name_fct
    this_operator = scope.declare_local_operator('fct_' + name)
    this_operator.raw_operator = op
    this_operator.inputs = operator.inputs

    score_name = scope.declare_local_variable(
        'scores', operator.inputs[0].type)
    this_operator.outputs.append(score_name)
    apply_identity(scope, score_name.full_name,
                   operator.outputs[0].full_name, container)


def empty_shape_calculator(operator):
    """
    Does nothing.
    """
    pass


def register_scorers():
    """
    Registers operators for @see cl CustomScorerTransform.
    """
    from .cdist_score import score_cdist_sum, convert_score_cdist_sum
    done = []
    update_registered_parser(
        CustomScorerTransform,
        custom_scorer_transform_parser)

    update_registered_converter(
        CustomScorerTransform,
        'MlprodictCustomScorerTransform',
        custom_scorer_transform_shape_calculator,
        custom_scorer_transform_converter)
    done.append(CustomScorerTransform)

    update_registered_converter(
        score_cdist_sum, 'fct_score_cdist_sum',
        empty_shape_calculator, convert_score_cdist_sum,
        options={'cdist': [None, 'single-node']})

    return done



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRunUnaryNum, OpRun


class Squeeze_1(OpRunUnaryNum):

    atts = {'axes': [], 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Squeeze_1.atts,
                               **options)
        if isinstance(self.axes, numpy.ndarray):
            self.axes = tuple(self.axes)
        elif self.axes in [[], tuple()]:
            self.axes = None
        elif isinstance(self.axes, list):
            self.axes = tuple(self.axes)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if isinstance(self.axes, (tuple, list)):
            sq = data
            for a in reversed(self.axes):
                sq = numpy.squeeze(sq, axis=a)
        else:
            sq = numpy.squeeze(data, axis=self.axes)
        return (sq, )


class Squeeze_11(Squeeze_1):
    pass


class Squeeze_13(OpRun):

    atts = {'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Squeeze_13.atts,
                       **options)
        self.axes = None

    def _run(self, data, axes=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if axes is not None:
            if hasattr(axes, '__iter__'):
                sq = numpy.squeeze(data, axis=tuple(axes))
            else:
                sq = numpy.squeeze(data, axis=axes)
        else:
            sq = numpy.squeeze(data)
        return (sq, )


if onnx_opset_version() >= 13:
    Squeeze = Squeeze_13
elif onnx_opset_version() >= 11:  # pragma: no cover
    Squeeze = Squeeze_11
else:  # pragma: no cover
    Squeeze = Squeeze_1



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Where(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, condition, x, y, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if x.dtype != y.dtype and x.dtype not in (numpy.object_, ):
            raise RuntimeError(  # pragma: no cover
                f"x and y should share the same dtype {x.dtype} != {y.dtype}")
        if x.shape != y.shape and x.shape != (1, ) and y.shape != (1, ):
            raise RuntimeError(  # pragma: no cover
                f"x and y should share the same shape {x.shape} != {y.shape}")
        return (numpy.where(condition, x, y).astype(x.dtype), )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from numpy.fft import fft
from ._op import OpRun
from ._new_ops import OperatorSchema


class FFT(OpRun):

    atts = {'axis': -1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=FFT.atts,
                       **options)

    def _find_custom_operator_schema(self, op_name):
        if op_name == "FFT":
            return FFTSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _run(self, a, fft_length=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if fft_length is not None:
            fft_length = fft_length[0]
            y = fft(a, fft_length, axis=self.axis)
        else:
            y = fft(a, axis=self.axis)
        if a.dtype in (numpy.float32, numpy.complex64):
            return (y.astype(numpy.complex64), )
        if a.dtype in (numpy.float64, numpy.complex128):
            return (y.astype(numpy.complex128), )
        raise TypeError(  # pragma: no cover
            f"Unexpected input type: {a.dtype!r}.")

    def to_python(self, inputs):
        if len(inputs) == 1:
            return ('from numpy.fft import fft',
                    f"return fft({inputs[0]}, axis={self.axis})")
        return ('from numpy.fft import fft',
                f"return fft({inputs[0]}, {inputs[1]}[0], axis={self.axis})")


class FFTSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl FFT.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'FFT')
        self.attributes = FFT.atts



"""
@file
@brief Helpers to manipulate :epkg:`scikit-learn` models.
"""
import inspect
import multiprocessing
import numpy
from sklearn.base import (
    TransformerMixin, ClassifierMixin, RegressorMixin, BaseEstimator)
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.compose import ColumnTransformer, TransformedTargetRegressor


def enumerate_pipeline_models(pipe, coor=None, vs=None):
    """
    Enumerates all the models within a pipeline.

    @param      pipe        *scikit-learn* pipeline
    @param      coor        current coordinate
    @param      vs          subset of variables for the model, None for all
    @return                 iterator on models ``tuple(coordinate, model)``

    Example:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from sklearn.datasets import load_iris
        from sklearn.decomposition import PCA
        from sklearn.linear_model import LogisticRegression
        from sklearn.pipeline import make_pipeline
        from sklearn.model_selection import train_test_split
        from mlprodict.onnx_tools.optim.sklearn_helper import enumerate_pipeline_models

        iris = load_iris()
        X, y = iris.data, iris.target
        X_train, __, y_train, _ = train_test_split(X, y, random_state=11)
        clr = make_pipeline(PCA(n_components=2),
                            LogisticRegression(solver="liblinear"))
        clr.fit(X_train, y_train)

        for a in enumerate_pipeline_models(clr):
            print(a)
    """
    if coor is None:
        coor = (0,)
    yield coor, pipe, vs
    if hasattr(pipe, 'transformer_and_mapper_list') and len(pipe.transformer_and_mapper_list):
        # azureml DataTransformer
        raise NotImplementedError(  # pragma: no cover
            "Unable to handle this specific case.")
    elif hasattr(pipe, 'mapper') and pipe.mapper:
        # azureml DataTransformer
        for couple in enumerate_pipeline_models(  # pragma: no cover
                pipe.mapper, coor + (0,)):
            yield couple
    elif hasattr(pipe, 'built_features'):  # pragma: no cover
        # sklearn_pandas.dataframe_mapper.DataFrameMapper
        for i, (columns, transformers, _) in enumerate(
                pipe.built_features):
            if isinstance(columns, str):
                columns = (columns,)
            if transformers is None:
                yield (coor + (i,)), None, columns
            else:
                for couple in enumerate_pipeline_models(transformers, coor + (i,), columns):
                    yield couple
    elif isinstance(pipe, Pipeline):
        for i, (_, model) in enumerate(pipe.steps):
            for couple in enumerate_pipeline_models(model, coor + (i,)):
                yield couple
    elif isinstance(pipe, ColumnTransformer):
        for i, (_, fitted_transformer, column) in enumerate(pipe.transformers):
            for couple in enumerate_pipeline_models(
                    fitted_transformer, coor + (i,), column):
                yield couple
    elif isinstance(pipe, FeatureUnion):
        for i, (_, model) in enumerate(pipe.transformer_list):
            for couple in enumerate_pipeline_models(model, coor + (i,)):
                yield couple
    elif isinstance(pipe, TransformedTargetRegressor):
        raise NotImplementedError(
            "Not yet implemented for TransformedTargetRegressor.")
    elif isinstance(pipe, (TransformerMixin, ClassifierMixin, RegressorMixin)):
        pass
    elif isinstance(pipe, BaseEstimator):
        pass
    elif isinstance(pipe, (list, numpy.ndarray)):
        for i, m in enumerate(pipe):
            for couple in enumerate_pipeline_models(m, coor + (i,)):
                yield couple
    else:
        raise TypeError(  # pragma: no cover
            f"pipe is not a scikit-learn object: {type(pipe)}\n{pipe}")


def enumerate_fitted_arrays(model):
    """
    Enumerate all fitted arrays included in a
    :epkg:`scikit-learn` object.

    @param      model       :epkg:`scikit-learn` object
    @return                 enumerator

    One example:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from sklearn.datasets import load_iris
        from sklearn.decomposition import PCA
        from sklearn.linear_model import LogisticRegression
        from sklearn.pipeline import make_pipeline
        from sklearn.model_selection import train_test_split
        from mlprodict.onnx_tools.optim.sklearn_helper import enumerate_fitted_arrays

        iris = load_iris()
        X, y = iris.data, iris.target
        X_train, __, y_train, _ = train_test_split(X, y, random_state=11)
        clr = make_pipeline(PCA(n_components=2),
                            LogisticRegression(solver="liblinear"))
        clr.fit(X_train, y_train)

        for a in enumerate_fitted_arrays(clr):
            print(a)
    """
    def enumerate__(obj):
        if isinstance(obj, (tuple, list)):
            for el in obj:
                for o in enumerate__(el):
                    yield (obj, el, o)
        elif isinstance(obj, dict):
            for k, v in obj.items():
                for o in enumerate__(v):
                    yield (obj, k, v, o)
        elif hasattr(obj, '__dict__'):
            for k, v in obj.__dict__.items():
                if k[-1] != '_' and k[0] != '_':
                    continue
                if isinstance(v, numpy.ndarray):
                    yield (obj, k, v)
                else:
                    for row in enumerate__(v):
                        yield row

    for row in enumerate_pipeline_models(model):
        coord = row[:-1]
        sub = row[1]
        last = row[2:]
        for sub_row in enumerate__(sub):
            yield coord + (sub, sub_row) + last


def pairwise_array_distances(l1, l2, metric='l1med'):
    """
    Computes pairwise distances between two lists of arrays
    *l1* and *l2*. The distance is 1e9 if shapes are not equal.

    @param      l1          first list of arrays
    @param      l2          second list of arrays
    @param      metric      metric to use, `'l1med'` compute
                            the average absolute error divided
                            by the ansolute median
    @return                 matrix
    """
    dist = numpy.full((len(l1), len(l2)), 1e9)
    for i, a1 in enumerate(l1):
        if not isinstance(a1, numpy.ndarray):
            continue  # pragma: no cover
        for j, a2 in enumerate(l2):
            if not isinstance(a2, numpy.ndarray):
                continue  # pragma: no cover
            if a1.shape != a2.shape:
                continue
            a = numpy.median(numpy.abs(a1))
            if a == 0:
                a = 1
            diff = numpy.sum(numpy.abs(a1 - a2)) / a
            dist[i, j] = diff / diff.size
    return dist


def max_depth(estimator):
    """
    Retrieves the max depth assuming the estimator
    is a decision tree.
    """
    n_nodes = estimator.tree_.node_count
    children_left = estimator.tree_.children_left
    children_right = estimator.tree_.children_right

    node_depth = numpy.zeros(shape=n_nodes, dtype=numpy.int64)
    is_leaves = numpy.zeros(shape=n_nodes, dtype=bool)
    stack = [(0, -1)]  # seed is the root node id and its parent depth
    while len(stack) > 0:
        node_id, parent_depth = stack.pop()
        node_depth[node_id] = parent_depth + 1

        # If we have a test node
        if children_left[node_id] != children_right[node_id]:
            stack.append((children_left[node_id], parent_depth + 1))
            stack.append((children_right[node_id], parent_depth + 1))
        else:
            is_leaves[node_id] = True
    return max(node_depth)


def inspect_sklearn_model(model, recursive=True):
    """
    Inspects a :epkg:`scikit-learn` model and produces
    some figures which tries to represent the complexity of it.

    @param      model       model
    @param      recursive   recursive look
    @return                 dictionary

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import pprint
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.linear_model import LogisticRegression
        from sklearn.datasets import load_iris
        from mlprodict.onnx_tools.optim.sklearn_helper import inspect_sklearn_model

        iris = load_iris()
        X = iris.data
        y = iris.target
        lr = LogisticRegression()
        lr.fit(X, y)
        pprint.pprint((lr, inspect_sklearn_model(lr)))


        iris = load_iris()
        X = iris.data
        y = iris.target
        rf = RandomForestClassifier()
        rf.fit(X, y)
        pprint.pprint((rf, inspect_sklearn_model(rf)))
    """
    def update(sts, st):
        for k, v in st.items():
            if k in sts:
                if 'max_' in k:
                    sts[k] = max(v, sts[k])
                else:
                    sts[k] += v
            else:
                sts[k] = v

    def insmodel(m):
        st = {'nop': 1}
        if hasattr(m, 'tree_') and hasattr(m.tree_, 'node_count'):
            st['nnodes'] = m.tree_.node_count
            st['ntrees'] = 1
            st['max_depth'] = max_depth(m)
        try:
            if hasattr(m, 'coef_'):
                st['ncoef'] = len(m.coef_)
                st['nlin'] = 1
        except KeyError:  # pragma: no cover
            # added to deal with xgboost 1.0 (KeyError: 'weight')
            pass
        if hasattr(m, 'estimators_'):
            for est in m.estimators_:
                st_ = inspect_sklearn_model(est, recursive=recursive)
                update(st, st_)
        return st

    if recursive:
        sts = {}
        for __, m, _ in enumerate_pipeline_models(model):
            st = inspect_sklearn_model(m, recursive=False)
            update(sts, st)
            st = insmodel(m)
            update(sts, st)
        return st
    return insmodel(model)


def set_n_jobs(model, params, n_jobs=None):
    """
    Looks into model signature and add parameter *n_jobs*
    if available. The function does not overwrite the parameter.

    @param      model       model class
    @param      params      current set of parameters
    @param      n_jobs      number of CPU or *n_jobs* if specified or 0
    @return                 new set of parameters

    On this machine, the default value is the following.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import multiprocessing
        print(multiprocessing.cpu_count())
    """
    if params is not None and 'n_jobs' in params:
        return params
    sig = inspect.signature(model.__init__)
    if 'n_jobs' not in sig.parameters:
        return params
    if n_jobs == 0:
        n_jobs = None
    params = params.copy() if params else {}
    params['n_jobs'] = n_jobs or multiprocessing.cpu_count()
    return params



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class Normalizer(OpRunUnaryNum):

    atts = {'norm': 'MAX'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Normalizer.atts,
                               **options)
        if self.norm == b'MAX':  # pylint: disable=E1101
            self._norm = Normalizer.norm_max
        elif self.norm == b'L1':  # pylint: disable=E1101
            self._norm = Normalizer.norm_l1
        elif self.norm == b'L2':  # pylint: disable=E1101
            self._norm = Normalizer.norm_l2
        else:
            raise ValueError(  # pragma: no cover
                f"Unexpected value for norm='{self.norm}'.")  # pylint: disable=E1101

    @staticmethod
    def norm_max(x, inplace):
        "max normalization"
        if inplace:
            return Normalizer._norm_max_inplace(x)
        return x / numpy.abs(x).max(axis=1).reshape((x.shape[0], -1))

    @staticmethod
    def _norm_max_inplace(x):
        numpy.divide(x, numpy.abs(x).max(axis=1).reshape((x.shape[0], -1)),
                     out=x)
        return x

    @staticmethod
    def norm_l1(x, inplace):
        "L1 normalization"
        if inplace:
            return Normalizer._norm_L1_inplace(x)
        return x / numpy.abs(x).sum(axis=1).reshape((x.shape[0], -1))

    @staticmethod
    def _norm_L1_inplace(x):
        numpy.divide(x, numpy.abs(x).sum(axis=1).reshape((x.shape[0], -1)),
                     out=x)
        return x

    @staticmethod
    def norm_l2(x, inplace):
        "L2 normalization"
        xn = numpy.square(x).sum(axis=1)
        numpy.sqrt(xn, out=xn)
        norm = xn.reshape((x.shape[0], -1))
        if inplace:
            numpy.divide(x, norm, out=x)
            return x
        return x / norm

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (self._norm(
            x, inplace=self.inplaces.get(0, False) and x.flags['WRITEABLE']), )



# -*- encoding: utf-8 -*-
# pylint: disable=C0302,R0914
"""
@file
@brief Overloads a conversion function.
"""
import json
import pprint
from collections import OrderedDict
import logging
import numpy
from onnx import ValueInfoProto
import pandas
try:
    from sklearn.metrics._scorer import _PredictScorer
except ImportError:  # pragma: no cover
    # scikit-learn < 0.22
    from sklearn.metrics.scorer import _PredictScorer
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.compose import ColumnTransformer
from sklearn.utils.metaestimators import _BaseComposition
from skl2onnx.common.data_types import (
    FloatTensorType, DoubleTensorType, DataType, guess_numpy_type,
    StringTensorType, Int64TensorType, _guess_type_proto)
from skl2onnx import convert_sklearn
from skl2onnx.algebra.onnx_operator_mixin import OnnxOperatorMixin
from skl2onnx.algebra.type_helper import _guess_type
from ..onnx_tools.onnx_manipulations import onnx_rename_names
from ..onnx_tools.onnx2py_helper import (
    guess_dtype, get_tensor_shape, get_tensor_elem_type)
from .register_rewritten_converters import register_rewritten_operators
from .register import register_converters
from .scorers import CustomScorerTransform


logger = logging.getLogger('mlprodict')


def _fix_opset_skl2onnx():
    import skl2onnx
    from .. import __max_supported_opset__
    if skl2onnx.__max_supported_opset__ != __max_supported_opset__:
        skl2onnx.__max_supported_opset__ = __max_supported_opset__  # pragma: no cover


def convert_scorer(fct, initial_types, name=None,
                   target_opset=None, options=None,
                   custom_conversion_functions=None,
                   custom_shape_calculators=None,
                   custom_parsers=None, white_op=None,
                   black_op=None, final_types=None,
                   verbose=0):
    """
    Converts a scorer into :epkg:`ONNX` assuming
    there exists a converter associated to it.
    The function wraps the function into a custom
    transformer, then calls function *convert_sklearn*
    from :epkg:`sklearn-onnx`.

    :param fct: function to convert (or a scorer from :epkg:`scikit-learn`)
    :param initial_types: types information
    :param name: name of the produced model
    :param target_opset: to do it with a different target opset
    :param options: additional parameters for the conversion
    :param custom_conversion_functions: a dictionary for specifying the user
        customized conversion function, it takes precedence over
        registered converters
    :param custom_shape_calculators: a dictionary for specifying the user
        customized shape calculator it takes precedence over registered
        shape calculators.
    :param custom_parsers: parsers determine which outputs is expected
        for which particular task, default parsers are
        defined for classifiers, regressors, pipeline but
        they can be rewritten, *custom_parsers* is a dictionary
        ``{ type: fct_parser(scope, model, inputs,
        custom_parsers=None) }``
    :param white_op: white list of ONNX nodes allowed
        while converting a pipeline, if empty, all are allowed
    :param black_op: black list of ONNX nodes allowed
        while converting a pipeline, if empty, none are blacklisted
    :param final_types: a python list. Works the same way as
        initial_types but not mandatory, it is used
        to overwrites the type (if type is not None)
        and the name of every output.
    :param verbose: displays information while converting
    :return: :epkg:`ONNX` graph
    """
    if hasattr(fct, '_score_func'):
        kwargs = fct._kwargs
        fct = fct._score_func
    else:
        kwargs = None  # pragma: no cover
    if name is None:
        name = f"mlprodict_fct_ONNX({fct.__name__})"
    tr = CustomScorerTransform(fct.__name__, fct, kwargs)
    _fix_opset_skl2onnx()
    return convert_sklearn(
        tr, initial_types=initial_types,
        target_opset=target_opset, options=options,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
        custom_parsers=custom_parsers, white_op=white_op,
        black_op=black_op, final_types=final_types,
        verbose=verbose)


def guess_initial_types(X, initial_types):
    """
    Guesses initial types from an array or a dataframe.

    :param X: array or dataframe
    :param initial_types: hints about X
    :return: data types
    """
    if X is None and initial_types is None:
        raise NotImplementedError(  # pragma: no cover
            "Initial types must be specified.")
    elif initial_types is None:
        if isinstance(X, (numpy.ndarray, pandas.DataFrame)):
            X = X[:1]
        if isinstance(X, pandas.DataFrame):
            initial_types = []
            for c in X.columns:
                if isinstance(X[c].values[0], (str, numpy.str_)):
                    g = StringTensorType()
                else:
                    g = _guess_type(X[c].values)
                g.shape = [None, 1]
                initial_types.append((c, g))
        else:
            gt = _guess_type(X)
            initial_types = [('X', gt)]
    return initial_types


def _replace_tensor_type(schema, tensor_type):
    res = []
    for name, ty in schema:
        cl = ty.__class__
        if cl in (FloatTensorType, DoubleTensorType) and cl != tensor_type:
            ty = tensor_type(ty.shape)
        res.append((name, ty))
    return res


def guess_schema_from_data(X, tensor_type=None, schema=None):
    """
    Guesses initial types from a dataset.

    @param      X               dataset (dataframe, array)
    @param      tensor_type     if not None, replaces every
                                *FloatTensorType* or *DoubleTensorType*
                                by this one
    @param      schema          known schema
    @return                     schema (list of typed and named columns)
    """
    init = guess_initial_types(X, schema)
    if tensor_type is not None:
        init = _replace_tensor_type(init, tensor_type)
    # Grouping column
    unique = set()
    for _, col in init:
        if len(col.shape) != 2:
            return init  # pragma: no cover
        if col.shape[0] is not None:
            return init  # pragma: no cover
        if len(unique) > 0 and col.__class__ not in unique:
            return init  # pragma: no cover
        unique.add(col.__class__)
    unique = list(unique)
    return [('X', unique[0]([None, sum(_[1].shape[1] for _ in init)]))]


def get_inputs_from_data(X, schema=None):
    """
    Produces input data for *onnx* runtime.

    @param  X       data
    @param  schema  schema if None, schema is guessed with
        @see fn guess_schema_from_data
    @return input data
    """
    def _cast_data(X, ct):
        if isinstance(ct, FloatTensorType):
            return X.astype(numpy.float32)
        if isinstance(ct, DoubleTensorType):
            return X.astype(numpy.float64)
        if isinstance(ct, StringTensorType):
            return X.astype(numpy.str_)
        if isinstance(ct, Int64TensorType):
            return X.astype(numpy.int64)
        raise RuntimeError(  # pragma: no cover
            f"Unexpected column type {ct} for type {type(X)}.")

    if schema is None:
        schema = guess_schema_from_data(X)
    if isinstance(X, numpy.ndarray):
        if len(schema) != 1:
            raise RuntimeError(  # pragma: no cover
                "More than one column but input is an array.")
        return {schema[0][0]: _cast_data(X, schema[0][1])}
    if isinstance(X, pandas.DataFrame):
        if len(schema) != X.shape[1]:
            raise RuntimeError(  # pragma: no cover
                "Mismatch between onnx columns {} and DataFrame columns {}"
                "".format(len(schema), X.shape[1]))
        return {sch[0]: _cast_data(X[c].values, sch[1]).reshape((-1, 1))
                for sch, c in zip(schema, X.columns)}
    raise TypeError(  # pragma: no cover
        f"Unexpected type {type(X)}, expecting an array or a dataframe.")


def guess_schema_from_model(model, tensor_type=None, schema=None):
    """
    Guesses initial types from a model.

    @param      model           model
    @param      tensor_type     if not None, replaces every
                                *FloatTensorType* or *DoubleTensorType*
                                by this one
    @param      schema          known schema
    @return                     schema (list of typed and named columns)
    """
    if schema is not None:
        try:
            guessed = guess_schema_from_model(model)
        except NotImplementedError:  # pragma: no cover
            return _replace_tensor_type(schema, tensor_type)
        if len(guessed) != len(schema):
            raise RuntimeError(  # pragma: no cover
                "Given schema and guessed schema are not the same:\nGOT: {}\n-----\nGOT:\n{}".format(
                    schema, guessed))
        return _replace_tensor_type(schema, tensor_type)

    if hasattr(model, 'coef_'):
        # linear model
        init = [('X', FloatTensorType([None, model.coef_.shape[1]]))]
        return _replace_tensor_type(init, tensor_type)
    elif hasattr(model, 'dump_model'):
        dumped = model.dump_model()
        if isinstance(dumped, dict) and 'feature_names' in dumped:
            names = dumped['feature_names']
            init = [(name, FloatTensorType([None, 1])) for name in names]
            return _replace_tensor_type(init, tensor_type)

    data = pprint.pformat(model.__dict__)
    dirs = pprint.pformat(dir(model))
    if hasattr(model, 'dump_model'):  # pragma: no cover
        dumped = model.dump_model()
        keys = list(sorted(dumped))
        last = pprint.pformat([keys, dumped])
        if len(last) >= 200000:
            last = last[:200000] + "\n..."
    else:
        last = ""
    raise NotImplementedError(  # pragma: no cover
        "Unable to guess schema for model {}\n{}\n----\n{}\n------\n{}".format(
            model.__class__, data, dirs, last))


def _guess_type_(X, itype, dtype):
    initial_types = guess_initial_types(X, itype)
    if dtype is None:
        if hasattr(X, 'dtypes'):  # DataFrame
            dtype = numpy.float32
        elif hasattr(X, 'dtype'):
            dtype = X.dtype
        elif hasattr(X, 'type'):
            dtype = guess_numpy_type(X.type)
        elif isinstance(initial_types[0], ValueInfoProto):
            dtype = guess_dtype(initial_types[0].type.tensor_type.elem_type)
        elif initial_types is not None:
            dtype = guess_numpy_type(initial_types[0][1])
        else:
            raise RuntimeError(  # pragma: no cover
                f"dtype cannot be guessed: {type(X)}")
        if dtype != numpy.float64:
            dtype = numpy.float32
    if dtype is None:
        raise RuntimeError("dtype cannot be None")  # pragma: no cover
    if isinstance(dtype, FloatTensorType):
        dtype = numpy.float32  # pragma: no cover
    elif isinstance(dtype, DoubleTensorType):
        dtype = numpy.float64  # pragma: no cover
    new_dtype = dtype
    if isinstance(dtype, numpy.ndarray):
        new_dtype = dtype.dtype  # pragma: no cover
    elif isinstance(dtype, DataType):
        new_dtype = numpy.float32  # pragma: no cover
    if new_dtype not in (numpy.float32, numpy.float64, numpy.int64,
                         numpy.int32, numpy.float16):
        raise NotImplementedError(  # pragma: no cover
            f"dtype should be real not {new_dtype} ({dtype})")
    return initial_types, dtype, new_dtype


def to_onnx(model, X=None, name=None, initial_types=None,
            target_opset=None, options=None, rewrite_ops=False,
            white_op=None, black_op=None, final_types=None,
            rename_strategy=None, verbose=0,
            as_function=False, prefix_name=None,
            run_shape=False, single_function=True):
    """
    Converts a model using on :epkg:`sklearn-onnx`.

    :param model: model to convert or a function
        wrapped into :epkg:`_PredictScorer` with
        function :epkg:`make_scorer`
    :param X: training set (at least one row),
        can be None, it is used to infered the
        input types (*initial_types*)
    :param initial_types: if *X* is None, then *initial_types*
        must be defined
    :param name: name of the produced model
    :param target_opset: to do it with a different target opset
    :param options: additional parameters for the conversion
    :param rewrite_ops: rewrites some existing converters,
        the changes are permanent
    :param white_op: white list of ONNX nodes allowed
        while converting a pipeline, if empty, all are allowed
    :param black_op: black list of ONNX nodes allowed
        while converting a pipeline, if empty,
        none are blacklisted
    :param final_types: a python list. Works the same way as
        initial_types but not mandatory, it is used
        to overwrites the type (if type is not None)
        and the name of every output.
    :param rename_strategy: rename any name in the graph, select shorter
        names, see @see fn onnx_rename_names
    :param verbose: display information while converting the model
    :param as_function: exposes every model in a pipeline as a function,
        the main graph contains the pipeline structure,
        see :ref:`onnxsklearnfunctionsrst` for an example
    :param prefix_name: used if *as_function* is True, to give
        a prefix to variable in a pipeline
    :param run_shape: run shape inference
    :param single_function: if *as_function* is True, the function returns one graph
        with one call to the main function if *single_function* is True or
        a list of node corresponding to the graph structure
    :return: converted model

    The function rewrites function *to_onnx* from :epkg:`sklearn-onnx`
    but may changes a few converters if *rewrite_ops* is True.
    For example, :epkg:`ONNX` only supports *TreeEnsembleRegressor*
    for float but not for double. It becomes available
    if ``rewrite_ops=True``.

    .. faqref::
        :title: How to deal with a dataframe as input?

        Each column of the dataframe is considered as an named input.
        The first step is to make sure that every column type is correct.
        :epkg:`pandas` tends to select the least generic type to
        hold the content of one column. :epkg:`ONNX` does not automatically
        cast the data it receives. The data must have the same type with
        the model is converted and when the converted model receives
        the data to predict.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            from io import StringIO
            from textwrap import dedent
            import numpy
            import pandas
            from pyquickhelper.pycode import ExtTestCase
            from sklearn.preprocessing import OneHotEncoder
            from sklearn.pipeline import Pipeline
            from sklearn.compose import ColumnTransformer
            from mlprodict.onnx_conv import to_onnx
            from mlprodict.onnxrt import OnnxInference

            text = dedent('''
                __SCHEMA__
                7.4,0.7,0.0,1.9,0.076,11.0,34.0,0.9978,3.51,0.56,9.4,5,red
                7.8,0.88,0.0,2.6,0.098,25.0,67.0,0.9968,3.2,0.68,9.8,5,red
                7.8,0.76,0.04,2.3,0.092,15.0,54.0,0.997,3.26,0.65,9.8,5,red
                11.2,0.28,0.56,1.9,0.075,17.0,60.0,0.998,3.16,0.58,9.8,6,red
                ''')
            text = text.replace(
                "__SCHEMA__",
                "fixed_acidity,volatile_acidity,citric_acid,residual_sugar,chlorides,"
                "free_sulfur_dioxide,total_sulfur_dioxide,density,pH,sulphates,"
                "alcohol,quality,color")

            X_train = pandas.read_csv(StringIO(text))
            for c in X_train.columns:
                if c != 'color':
                    X_train[c] = X_train[c].astype(numpy.float32)
            numeric_features = [c for c in X_train if c != 'color']

            pipe = Pipeline([
                ("prep", ColumnTransformer([
                    ("color", Pipeline([
                        ('one', OneHotEncoder()),
                        ('select', ColumnTransformer(
                            [('sel1', 'passthrough', [0])]))
                    ]), ['color']),
                    ("others", "passthrough", numeric_features)
                ])),
            ])

            pipe.fit(X_train)
            pred = pipe.transform(X_train)
            print(pred)

            model_onnx = to_onnx(pipe, X_train, target_opset=12)
            oinf = OnnxInference(model_onnx)

            # The dataframe is converted into a dictionary,
            # each key is a column name, each value is a numpy array.
            inputs = {c: X_train[c].values for c in X_train.columns}
            inputs = {c: v.reshape((v.shape[0], 1)) for c, v in inputs.items()}

            onxp = oinf.run(inputs)
            print(onxp)

    .. versionchanged:: 0.9
        Parameter *as_function* was added.
    """
    logger.debug("to_onnx(%s, X=%r, initial_types=%r, target_opset=%r, "
                 "options=%r, rewrite_ops=%r, white_op=%r, black_op=%r, "
                 "final_types=%r)",
                 model.__class__.__name__, type(X), initial_types,
                 target_opset, options, rewrite_ops, white_op, black_op,
                 final_types)

    if isinstance(model, OnnxOperatorMixin):
        if not hasattr(model, 'op_version'):
            raise RuntimeError(  # pragma: no cover
                f"Missing attribute 'op_version' for type '{type(model)}'.")
        _fix_opset_skl2onnx()
        return model.to_onnx(
            X=X, name=name, options=options, black_op=black_op,
            white_op=white_op, final_types=final_types,
            target_opset=target_opset)
        # verbose=verbose)

    if rewrite_ops:
        old_values, old_shapes = register_rewritten_operators()
        register_converters()
    else:
        old_values, old_shapes = {}, {}

    if as_function and isinstance(
            model, (ColumnTransformer, Pipeline, FeatureUnion)):
        res = to_onnx_function(
            model, X=X, name=name, initial_types=initial_types,
            target_opset=target_opset, options=options,
            rewrite_ops=False,  # already handled
            white_op=white_op, black_op=black_op, final_types=final_types,
            rename_strategy=None,  # already handled
            verbose=verbose, prefix_name=prefix_name,
            run_shape=run_shape, single_function=single_function)

    elif isinstance(model, _PredictScorer):
        if X is not None and not isinstance(X, OrderedDict):
            raise ValueError("For a scorer, parameter X should be a OrderedDict not {}."
                             "".format(type(X)))
        if initial_types is None:
            dts = []
            initial_types = []
            for k, v in X.items():
                if hasattr(v, 'dtype'):
                    dtype = guess_numpy_type(v.dtype)
                else:
                    dtype = v  # pragma: no cover
                it, _, ndt = _guess_type_(v, None, dtype)
                for i in range(len(it)):  # pylint: disable=C0200
                    it[i] = (k, it[i][1])  # pylint: disable=C0200
                initial_types.extend(it)
                dts.append(ndt)
            ndt = set(dts)
            if len(ndt) != 1:
                raise RuntimeError(  # pragma: no cover
                    f"Multiple dtype is not efficient {ndt}.")
        res = convert_scorer(model, initial_types, name=name,
                             target_opset=target_opset, options=options,
                             black_op=black_op, white_op=white_op,
                             final_types=final_types, verbose=verbose)
    else:
        if name is None:
            name = f"mlprodict_ONNX({model.__class__.__name__})"

        initial_types, dtype, _ = _guess_type_(X, initial_types, None)

        _fix_opset_skl2onnx()
        res = convert_sklearn(model, initial_types=initial_types, name=name,
                              target_opset=target_opset, options=options,
                              black_op=black_op, white_op=white_op,
                              final_types=final_types, verbose=verbose)

    register_rewritten_operators(old_values, old_shapes)

    # optimisation
    if rename_strategy is not None:
        res = onnx_rename_names(res, strategy=rename_strategy)
    return res


def _guess_s2o_type(vtype: ValueInfoProto):
    return _guess_type_proto(
        get_tensor_elem_type(vtype), get_tensor_shape(vtype))


def _new_options(options, prefix, sklop):
    if sklop is None:
        raise RuntimeError(  # pragma: no cover
            "sklop cannot be None.")
    if isinstance(sklop, str):
        return None  # pragma: no cover
    if options is None:
        step_options = None
    else:
        step_options = {}
        for k, v in options.items():
            if k.startswith(prefix):
                step_options[k[len(prefix):]] = v
            elif '__' in k:
                step_options[k.split('__', maxsplit=1)[1]] = v
            if isinstance(sklop, _BaseComposition):
                step_options[k] = v
            else:
                from skl2onnx._supported_operators import _get_sklearn_operator_name
                from skl2onnx.common._registration import get_converter
                alias = _get_sklearn_operator_name(type(sklop))
                if alias is None:
                    step_options[k] = v
                else:
                    conv = get_converter(alias)
                    allowed = conv.get_allowed_options()
                    if allowed is not None and k in allowed:
                        step_options[k] = v
    return step_options


class _ParamEncoder(json.JSONEncoder):
    def default(self, obj):  # pylint: disable=W0237
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError as e:
            # Unable to serialize
            return '{"classname": "%s", "EXC": "%s"}' % (
                obj.__class__.__name__, str(e))


def get_sklearn_json_params(model):
    """
    Retrieves all the parameters of a :epkg:`scikit-learn` model.
    """
    pars = model.get_params(deep=False)
    try:
        return json.dumps(pars, cls=_ParamEncoder)
    except TypeError as e:  # pragma: no cover
        raise RuntimeError(
            f"Unable to serialize parameters {pprint.pformat(pars)}.") from e


def _to_onnx_function_pipeline(
        model, X=None, name=None, initial_types=None,
        target_opset=None, options=None, rewrite_ops=False,
        white_op=None, black_op=None, final_types=None,
        rename_strategy=None, verbose=0,
        prefix_name=None, run_shape=False,
        single_function=True):

    from ..npy.xop_variable import Variable
    from ..npy.xop import OnnxOperatorFunction, loadop
    from ..onnx_tools.onnx_manipulations import onnx_model_to_function

    OnnxIdentity = loadop('Identity')

    if len(model.steps) == 0:
        raise RuntimeError(  # pragma: no cover
            "The pipeline to be converted cannot be empty.")

    if target_opset is None:
        from .. import __max_supported_opset__
        op_version = __max_supported_opset__
    elif isinstance(target_opset, int):
        op_version = target_opset
    else:
        from .. import __max_supported_opset__
        op_version = target_opset.get('', __max_supported_opset__)

    i_types = guess_initial_types(X, initial_types)
    input_nodes = [OnnxIdentity(i[0], op_version=op_version)
                   for i in i_types]

    inputs = i_types
    last_op = None
    for i_step, step in enumerate(model.steps):
        prefix = step[0] + "__"
        step_options = _new_options(options, prefix, step[1])
        if prefix_name is not None:
            prefix = prefix_name + prefix
        protom = to_onnx(
            step[1], name=name, initial_types=inputs,
            target_opset=target_opset,
            options=step_options, rewrite_ops=rewrite_ops,
            white_op=white_op, black_op=black_op, verbose=verbose,
            as_function=True, prefix_name=prefix, run_shape=run_shape,
            single_function=False)
        for o in protom.graph.output:
            if get_tensor_elem_type(o) == 0:
                raise RuntimeError(  # pragma: no cover
                    "Unabble to guess output type of output %r "
                    "from model step %d: %r, output=%r." % (
                        protom.graph.output, i_step, step[1], o))
        jspar = 'HYPER:{"%s":%s}' % (
            step[1].__class__.__name__, get_sklearn_json_params(step[1]))
        protof, subf = onnx_model_to_function(
            protom, domain='sklearn',
            name=f"{prefix}_{step[1].__class__.__name__}_{i_step}",
            doc_string=jspar)
        input_names = [f"{step[0]}_{o}" for o in protof.input]
        if last_op is not None:
            if len(input_names) == 1:
                input_nodes = [OnnxIdentity(
                    last_op, output_names=input_names[0],
                    op_version=op_version)]
            else:
                input_nodes = [OnnxIdentity(last_op[i], output_names=[n],  # pylint: disable=E1136
                                            op_version=op_version)
                               for i, n in enumerate(input_names)]
        output_names = [f"{step[0]}_{o}" for o in protof.output]

        logger.debug("_to_onnx_function_pipeline:%s:%r->%r:%r:%s",
                     step[1].__class__.__name__,
                     input_names, output_names,
                     len(protof.node), jspar)

        op = OnnxOperatorFunction(
            protof, *input_nodes, output_names=output_names,
            sub_functions=subf)
        last_op = op
        inputs = [
            ('X%d' % i, _guess_s2o_type(o))
            for i, o in enumerate(protom.graph.output)]

    logger.debug("_to_onnx_function_pipeline:end:(%s-%d, X=%r, "
                 "initial_types=%r, target_opset=%r, "
                 "options=%r, rewrite_ops=%r, white_op=%r, black_op=%r, "
                 "final_types=%r, outputs=%r)",
                 model.__class__.__name__, id(model),
                 type(X), initial_types,
                 target_opset, options, rewrite_ops, white_op, black_op,
                 final_types, inputs)

    i_vars = [Variable.from_skl2onnx_tuple(i) for i in i_types]
    if final_types is None:
        outputs_tuple = [
            (n, _guess_s2o_type(o))
            for i, (n, o) in enumerate(zip(output_names, protom.graph.output))]
        outputs = [Variable.from_skl2onnx_tuple(i) for i in outputs_tuple]
    else:
        outputs = final_types

    onx = last_op.to_onnx(inputs=i_vars, target_opset=target_opset,
                          verbose=verbose, run_shape=run_shape,
                          outputs=outputs)

    for o in onx.graph.output:
        if get_tensor_elem_type(o) == 0:
            raise RuntimeError(  # pragma: no cover
                "Unable to guess output type of output %r "
                "from model %r." % (onx.graph.output, model))
    return onx


def get_column_index(i, inputs):
    """
    Returns a tuples (variable index, column index in that variable).
    The function has two different behaviours, one when *i* (column index)
    is an integer, another one when *i* is a string (column name).
    If *i* is a string, the function looks for input name with
    this name and returns `(index, 0)`.
    If *i* is an integer, let's assume first we have two inputs
    `I0 = FloatTensorType([None, 2])` and `I1 = FloatTensorType([None, 3])`,
    in this case, here are the results:

    ::

        get_column_index(0, inputs) -> (0, 0)
        get_column_index(1, inputs) -> (0, 1)
        get_column_index(2, inputs) -> (1, 0)
        get_column_index(3, inputs) -> (1, 1)
        get_column_index(4, inputs) -> (1, 2)
    """
    if isinstance(i, int):
        if i == 0:
            # Useful shortcut, skips the case when end is None
            # (unknown dimension)
            return 0, 0
        vi = 0
        pos = 0
        end = inputs[0][1].shape[1]
        if end is None:
            raise RuntimeError(  # pragma: no cover
                "Cannot extract a specific column %r when "
                "one input (%r) has unknown "
                "dimension." % (i, inputs[0]))
        while True:
            if pos <= i < end:
                return vi, i - pos
            vi += 1
            pos = end
            if vi >= len(inputs):
                raise RuntimeError(
                    "Input %r (i=%r, end=%r) is not available in\n%r" % (
                        vi, i, end, pprint.pformat(inputs)))
            rel_end = inputs[vi][1].shape[1]
            if rel_end is None:
                raise RuntimeError(  # pragma: no cover
                    "Cannot extract a specific column %r when "
                    "one input (%r) has unknown "
                    "dimension." % (i, inputs[vi]))
            end += rel_end
    else:
        for ind, inp in enumerate(inputs):
            if inp[0] == i:
                return ind, 0
        raise RuntimeError(  # pragma: no cover
            "Unable to find column name %r among names %r. "
            "Make sure the input names specified with parameter "
            "initial_types fits the column names specified in the "
            "pipeline to convert. This may happen because a "
            "ColumnTransformer follows a transformer without "
            "any mapped converter in a pipeline." % (
                i, [n[0] for n in inputs]))


def get_column_indices(indices, inputs, multiple):
    """
    Returns the requested graph inpudes based on their
    indices or names. See :func:`get_column_index`.

    :param indices: variables indices or names
    :param inputs: graph inputs
    :param multiple: allows column to come from multiple variables
    :return: a tuple *(variable name, list of requested indices)* if
        *multiple* is False, a dictionary *{ var_index: [ list of
        requested indices ] }*
        if *multiple* is True
    """
    if multiple:
        res = OrderedDict()
        for p in indices:
            ov, onnx_i = get_column_index(p, inputs)
            if ov not in res:
                res[ov] = []
            res[ov].append(onnx_i)
        return res

    onnx_var = None
    onnx_is = []
    for p in indices:
        ov, onnx_i = get_column_index(p, inputs)
        onnx_is.append(onnx_i)
        if onnx_var is None:
            onnx_var = ov
        elif onnx_var != ov:
            cols = [onnx_var, ov]
            raise NotImplementedError(  # pragma: no cover
                "sklearn-onnx is not able to merge multiple columns from "
                "multiple variables ({0}). You should think about merging "
                "initial types.".format(cols))
    return onnx_var, onnx_is


def _merge_initial_types(i_types, transform_inputs, merge):
    if len(i_types) == len(transform_inputs):
        new_types = []
        for it, sli in zip(i_types, transform_inputs):
            name, ty = it
            begin, end = sli.inputs[1], sli.inputs[2]
            delta = end - begin
            shape = [ty.shape[0], int(delta[0])]
            new_types.append((name, ty.__class__(shape)))
    else:
        raise NotImplementedError(  # pragma: no cover
            "Not implemented when i_types=%r, transform_inputs=%r."
            "" % (i_types, transform_inputs))
    if merge and len(new_types) > 1:
        raise NotImplementedError(  # pragma: no cover
            "Cannot merge %r built from i_types=%r, transform_inputs=%r."
            "" % (new_types, i_types, transform_inputs))
    return new_types


def _to_onnx_function_column_transformer(
        model, X=None, name=None, initial_types=None,
        target_opset=None, options=None, rewrite_ops=False,
        white_op=None, black_op=None, final_types=None,
        rename_strategy=None, verbose=0,
        prefix_name=None, run_shape=False,
        single_function=True):

    from sklearn.preprocessing import OneHotEncoder
    from ..npy.xop_variable import Variable
    from ..npy.xop import OnnxOperatorFunction, loadop
    from ..onnx_tools.onnx_manipulations import onnx_model_to_function

    OnnxConcat, OnnxSlice, OnnxIdentity = loadop('Concat', 'Slice', 'Identity')

    transformers = model.transformers_
    if len(transformers) == 0:
        raise RuntimeError(  # pragma: no cover
            "The ColumnTransformer to be converted cannot be empty.")

    if target_opset is None:
        from .. import __max_supported_opset__
        op_version = __max_supported_opset__
    elif isinstance(target_opset, int):
        op_version = target_opset
    else:  # pragma: no cover
        from .. import __max_supported_opset__
        op_version = target_opset.get('', __max_supported_opset__)

    i_types = guess_initial_types(X, initial_types)
    ops = []
    protoms = []
    output_namess = []
    for i_step, (name_step, op, column_indices) in enumerate(transformers):
        if op == 'drop':
            continue
        input_nodes = [OnnxIdentity(i[0], op_version=op_version)
                       for i in initial_types]
        if isinstance(column_indices, slice):
            column_indices = list(range(
                column_indices.start
                if column_indices.start is not None else 0,
                column_indices.stop, column_indices.step
                if column_indices.step is not None else 1))
        elif isinstance(column_indices, (int, str)):
            column_indices = [column_indices]
        names = get_column_indices(column_indices, i_types, multiple=True)
        transform_inputs = []
        for onnx_var, onnx_is in names.items():
            if max(onnx_is) - min(onnx_is) != len(onnx_is) - 1:
                raise RuntimeError(  # pragma: no cover
                    "The converter only with contiguous columns indices not %r "
                    "for step %r." % (column_indices, name_step))
            tr_inputs = OnnxSlice(input_nodes[onnx_var],
                                  numpy.array([onnx_is[0]], dtype=numpy.int64),
                                  numpy.array([onnx_is[-1] + 1],
                                              dtype=numpy.int64),
                                  numpy.array([1], dtype=numpy.int64),
                                  op_version=op_version)
            transform_inputs.append(tr_inputs)

        merged_cols = False
        if len(transform_inputs) > 1:
            if isinstance(op, Pipeline):
                if not isinstance(op.steps[0][1],
                                  (OneHotEncoder, ColumnTransformer)):
                    merged_cols = True
            elif not isinstance(op, (OneHotEncoder, ColumnTransformer)):
                merged_cols = True

        if merged_cols:
            concatenated = OnnxConcat(
                *transform_inputs, op_version=op_version, axis=1)
        else:
            concatenated = transform_inputs
        initial_types = _merge_initial_types(
            i_types, transform_inputs, merged_cols)

        prefix = name_step + "__"
        step_options = _new_options(options, prefix, op)
        if prefix_name is not None:
            prefix = prefix_name + prefix

        if op == 'passthrough':
            ops.extend(concatenated)
            continue

        protom = to_onnx(
            op, name=name_step, X=X, initial_types=initial_types,
            target_opset=target_opset,
            options=step_options, rewrite_ops=rewrite_ops,
            white_op=white_op, black_op=black_op, verbose=verbose,
            as_function=True, prefix_name=prefix, run_shape=run_shape,
            single_function=False)
        protoms.append(protom)

        for o in protom.graph.output:
            if get_tensor_elem_type(o) == 0:
                raise RuntimeError(  # pragma: no cover
                    "Unabble to guess output type of output %r "
                    "from model step %d: %r." % (
                        protom.graph.output, i_step, op))
        jspar = 'HYPER:{"%s":%s}' % (
            op.__class__.__name__, get_sklearn_json_params(op))
        protof, fcts = onnx_model_to_function(
            protom, domain='sklearn',
            name=f"{prefix}_{op.__class__.__name__}_{id(op)}",
            doc_string=jspar)
        output_names = [f"{name_step}_{o}" for o in protof.output]
        output_namess.append(output_names)

        logger.debug("_to_onnx_function_column_transformer:%s:->%r:%r:%s",
                     op.__class__.__name__, output_names, len(protof.node), jspar)

        op = OnnxOperatorFunction(
            protof, *concatenated, output_names=output_names,
            sub_functions=list(fcts))
        ops.append(op)

    logger.debug("_to_onnx_function_column_transformer:end:(%s-%d, X=%r, "
                 "initial_types=%r, target_opset=%r, "
                 "options=%r, rewrite_ops=%r, white_op=%r, black_op=%r, "
                 "final_types=%r, outputs=%r)",
                 model.__class__.__name__, id(model),
                 type(X), initial_types, target_opset,
                 options, rewrite_ops, white_op, black_op,
                 final_types, i_types)

    i_vars = [Variable.from_skl2onnx_tuple(i) for i in i_types]
    if final_types is None:
        outputs_tuple = []
        for protom, output_names in zip(protoms, output_namess):
            outputs_tuple.extend([
                (n, _guess_s2o_type(o))
                for i, (n, o) in enumerate(zip(output_names, protom.graph.output))])
        outputs = [Variable.from_skl2onnx_tuple(i) for i in outputs_tuple]
    else:
        outputs = final_types

    last_op = OnnxConcat(*ops, op_version=op_version, axis=1)

    onx = last_op.to_onnx(inputs=i_vars, target_opset=target_opset,
                          verbose=verbose, run_shape=run_shape,
                          outputs=outputs)

    for o in onx.graph.output:
        if get_tensor_elem_type(o) == 0:
            raise RuntimeError(  # pragma: no cover
                "Unable to guess output type of output %r "
                "from model %r." % (onx.graph.output, model))
    return onx


def to_onnx_function(model, X=None, name=None, initial_types=None,
                     target_opset=None, options=None, rewrite_ops=False,
                     white_op=None, black_op=None, final_types=None,
                     rename_strategy=None, verbose=0,
                     prefix_name=None, run_shape=False,
                     single_function=True):
    """
    Converts a model using on :epkg:`sklearn-onnx`.
    The functions works as the same as function @see fn to_onnx
    but every model is exported as a single function and the main
    graph represents the pipeline structure.

    :param model: model to convert or a function
        wrapped into :epkg:`_PredictScorer` with
        function :epkg:`make_scorer`
    :param X: training set (at least one row),
        can be None, it is used to infered the
        input types (*initial_types*)
    :param initial_types: if *X* is None, then *initial_types*
        must be defined
    :param name: name of the produced model
    :param target_opset: to do it with a different target opset
    :param options: additional parameters for the conversion
    :param rewrite_ops: rewrites some existing converters,
        the changes are permanent
    :param white_op: white list of ONNX nodes allowed
        while converting a pipeline, if empty, all are allowed
    :param black_op: black list of ONNX nodes allowed
        while converting a pipeline, if empty,
        none are blacklisted
    :param final_types: a python list. Works the same way as
        initial_types but not mandatory, it is used
        to overwrites the type (if type is not None)
        and the name of every output.
    :param rename_strategy: rename any name in the graph, select shorter
        names, see @see fn onnx_rename_names
    :param verbose: display information while converting the model
    :param prefix_name: prefix for variable names
    :param run_shape: run shape inference on the final onnx model
    :param single_function: if True, the main graph only includes one node
        calling the main function
    :return: converted model
    """
    if rename_strategy is not None or rewrite_ops:
        return to_onnx(
            model, X=X, name=name, initial_types=initial_types,
            target_opset=target_opset, options=options, rewrite_ops=rewrite_ops,
            white_op=white_op, black_op=black_op, final_types=final_types,
            rename_strategy=rename_strategy, verbose=verbose,
            run_shape=run_shape)

    logger.debug("to_onnx_function:begin:(%s-%d, X=%r, initial_types=%r, target_opset=%r, "
                 "options=%r, rewrite_ops=%r, white_op=%r, black_op=%r, "
                 "final_types=%r)",
                 model.__class__.__name__, id(model), type(X), initial_types,
                 target_opset, options, rewrite_ops, white_op, black_op,
                 final_types)

    if final_types is not None:
        raise NotImplementedError(  # pragma: no cover
            "final_types != None, not implemented yet.")

    if single_function and (not isinstance(model, Pipeline) or
                            len(model.steps) != 1):
        # Wraps the model into a single pipeline.
        new_model = Pipeline(steps=[('main', model)])
        return to_onnx_function(
            new_model, X=X, name=name, initial_types=initial_types,
            target_opset=target_opset, options=options, rewrite_ops=rewrite_ops,
            white_op=white_op, black_op=black_op, final_types=final_types,
            rename_strategy=rename_strategy, verbose=verbose,
            prefix_name=prefix_name, run_shape=run_shape, single_function=False)

    if isinstance(model, Pipeline):
        return _to_onnx_function_pipeline(
            model, X=X, name=name, initial_types=initial_types,
            target_opset=target_opset, options=options, rewrite_ops=rewrite_ops,
            white_op=white_op, black_op=black_op, final_types=final_types,
            rename_strategy=rename_strategy, verbose=verbose,
            prefix_name=prefix_name, run_shape=run_shape,
            single_function=single_function)

    if isinstance(model, ColumnTransformer):
        return _to_onnx_function_column_transformer(
            model, X=X, name=name, initial_types=initial_types,
            target_opset=target_opset, options=options, rewrite_ops=rewrite_ops,
            white_op=white_op, black_op=black_op, final_types=final_types,
            rename_strategy=rename_strategy, verbose=verbose,
            prefix_name=prefix_name, run_shape=run_shape,
            single_function=single_function)

    raise TypeError(  # pragma: no cover
        f"Unexpected type {type(model)!r} for model to convert.")



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from .op_softmax import Softmax


class LogSoftmax(Softmax):

    atts = {'axis': 1}

    def __init__(self, onnx_node, desc=None, **options):
        Softmax.__init__(self, onnx_node, desc=desc,
                         **options)

    def _run(self, X, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and X.flags['WRITEABLE']:
            return self._run_inplace(X)
        Y = Softmax._run(self, X)[0]
        numpy.log(Y, out=Y)
        return (Y, )

    def _run_inplace(self, X):
        Y = Softmax._run_inplace(self, X)[0]
        numpy.log(Y, out=Y)
        return (Y, )

    def to_python(self, inputs):
        lines = [
            "Y = {0} - {0}.max(axis=axis)[:, numpy.newaxis]".format(inputs[0]),
            "numpy.exp(Y, out=Y)",
            "Y /= Y.sum(axis=axis)[:, numpy.newaxis]",
            'numpy.log(Y, out=Y)',
            "return Y"]
        return ("import numpy", "\n".join(lines))



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_whole*.
"""
import json
import numpy


class OnnxWholeSession:
    """
    Runs the prediction for a single :epkg:`ONNX`,
    it lets the runtime handle the graph logic as well.

    :param onnx_data: :epkg:`ONNX` model or data
    :param runtime: runtime to be used, mostly :epkg:`onnxruntime`
    :param runtime_options: runtime options
    :param device: device, a string `cpu`, `cuda`, `cuda:0`...

    .. versionchanged:: 0.8
        Parameter *device* was added.
    """

    def __init__(self, onnx_data, runtime, runtime_options=None, device=None):
        if runtime not in ('onnxruntime1', 'onnxruntime1-cuda'):
            raise NotImplementedError(  # pragma: no cover
                f"runtime '{runtime}' is not implemented.")

        from onnxruntime import (  # delayed
            InferenceSession, SessionOptions, RunOptions,
            GraphOptimizationLevel)
        from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
            Fail as OrtFail, InvalidGraph as OrtInvalidGraph,
            InvalidArgument as OrtInvalidArgument,
            NotImplemented as OrtNotImplemented,
            RuntimeException as OrtRuntimeException)

        onnx_data0 = onnx_data
        if hasattr(onnx_data, 'SerializeToString'):
            onnx_data = onnx_data.SerializeToString()
        if isinstance(runtime_options, SessionOptions):
            sess_options = runtime_options
            session_options = None
            runtime_options = None
        else:
            session_options = (
                None if runtime_options is None
                else runtime_options.get('session_options', None))
            self.runtime = runtime
            sess_options = session_options or SessionOptions()
        self.run_options = RunOptions()
        self.run_options.log_severity_level = 3
        self.run_options.log_verbosity_level = 1

        if session_options is None:
            if runtime_options is not None:
                if runtime_options.get('disable_optimisation', False):
                    sess_options.graph_optimization_level = (  # pragma: no cover
                        GraphOptimizationLevel.ORT_ENABLE_ALL)
                if runtime_options.get('enable_profiling', True):
                    sess_options.enable_profiling = True
                if runtime_options.get('log_severity_level', 2) != 2:
                    v = runtime_options.get('log_severity_level', 2)
                    sess_options.log_severity_level = v
                    self.run_options.log_severity_level = v
        elif runtime_options is not None and 'enable_profiling' in runtime_options:
            raise RuntimeError(  # pragma: no cover
                "session_options and enable_profiling cannot be defined at the "
                "same time.")
        elif runtime_options is not None and 'disable_optimisation' in runtime_options:
            raise RuntimeError(  # pragma: no cover
                "session_options and disable_optimisation cannot be defined at the "
                "same time.")
        elif runtime_options is not None and 'log_severity_level' in runtime_options:
            raise RuntimeError(  # pragma: no cover
                "session_options and log_severity_level cannot be defined at the "
                "same time.")
        providers = ['CPUExecutionProvider']
        if runtime == 'onnxruntime1-cuda':
            providers = ['CUDAExecutionProvider'] + providers
        try:
            self.sess = InferenceSession(onnx_data, sess_options=sess_options,
                                         device=device, providers=providers)
        except (OrtFail, OrtNotImplemented, OrtInvalidGraph,
                OrtInvalidArgument, OrtRuntimeException, RuntimeError) as e:
            from ...plotting.text_plot import onnx_simple_text_plot
            raise RuntimeError(
                "Unable to create InferenceSession due to '{}'\n{}.".format(
                    e, onnx_simple_text_plot(onnx_data0, recursive=True))) from e
        self.output_names = [_.name for _ in self.sess.get_outputs()]

    def run(self, inputs):
        """
        Computes the predictions.

        @param      inputs      dictionary *{variable, value}*
        @return                 list of outputs
        """
        v = next(iter(inputs.values()))
        if isinstance(v, (numpy.ndarray, dict)):
            try:
                return self.sess._sess.run(
                    self.output_names, inputs, self.run_options)
            except ValueError as e:
                raise ValueError(
                    "Issue running inference inputs=%r, expected inputs=%r."
                    "" % (
                        list(sorted(inputs)),
                        [i.name for i in self.sess.get_inputs()])) from e
        try:
            return self.sess._sess.run_with_ort_values(
                inputs, self.output_names, self.run_options)
        except RuntimeError:
            return self.sess._sess.run_with_ort_values(
                {k: v._get_c_value() for k, v in inputs.items()},
                self.output_names, self.run_options)

    @staticmethod
    def process_profiling(js):
        """
        Flattens json returned by onnxruntime profiling.

        :param js: json
        :return: list of dictionaries
        """
        rows = []
        for row in js:
            if 'args' in row and isinstance(row['args'], dict):
                for k, v in row['args'].items():
                    row[f'args_{k}'] = v
                del row['args']
            rows.append(row)
        return rows

    def get_profiling(self):
        """
        Returns the profiling informations.
        """
        prof = self.sess.end_profiling()
        with open(prof, 'r') as f:
            content = f.read()
        js = json.loads(content)
        return OnnxWholeSession.process_profiling(js)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from .op_conv_transpose_ import (  # pylint: disable=E0611,E0401
    ConvTransposeFloat, ConvTransposeDouble)


class ConvTranspose(OpRun):

    atts = {'auto_pad': 'NOTSET', 'group': 1,
            'dilations': [],
            'kernel_shape': [],
            'pads': [],
            'strides': [],
            'output_padding': [],
            'output_shape': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ConvTranspose.atts,
                       **options)
        self._init()

    def _init(self):
        self.rt32_ = ConvTransposeFloat()
        self.rt64_ = ConvTransposeDouble()
        for rt in [self.rt32_, self.rt64_]:
            rt.init(self.auto_pad,
                    numpy.array(self.dilations, dtype=numpy.int64),
                    self.group,
                    numpy.array(self.kernel_shape, dtype=numpy.int64),
                    numpy.array(self.pads, dtype=numpy.int64),
                    numpy.array(self.strides, dtype=numpy.int64),
                    numpy.array(self.output_padding, dtype=numpy.int64),
                    numpy.array(self.output_shape, dtype=numpy.int64))

    def _run(self, X, W, B=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if X.dtype == numpy.float32:
            return (self.rt32_.compute(X, W, B), )
        return (self.rt64_.compute(X, W, B), )



"""
@file
@brief Micro runtime for ONNX.

.. versionadded:: 0.6
"""
import numpy
from ..onnx_tools.onnx2py_helper import _var_as_dict


class OnnxMicroRuntime:
    """
    Implements a micro runtime for ONNX graphs.
    It does not implements all the operator types.

    :param model_onnx: ONNX model

    .. runpython::
        :showcode:

        import pprint
        import numpy
        from mlprodict.onnxrt.onnx_micro_runtime import OnnxMicroRuntime
        from mlprodict.npy.xop import loadop

        OnnxAdd = loadop('Add')

        dtype = numpy.float32
        opset = 15
        x = numpy.array([1, 2, 4, 5, 5, 4]).astype(
            numpy.float32).reshape((3, 2))
        cop = OnnxAdd('X', numpy.array([1], dtype=dtype), op_version=opset)
        cop4 = OnnxAdd(cop, numpy.array([2], dtype=dtype), op_version=opset,
                       output_names=['Y'])
        model_def = cop4.to_onnx({'X': x}, target_opset=opset)
        rt = OnnxMicroRuntime(model_def)
        out = rt.run({'X': x})
        pprint.pprint(out)
    """

    def __init__(self, model_onnx):
        if not hasattr(model_onnx, 'graph'):
            raise TypeError(
                f"model_onnx is not an ONNX graph but {type(model_onnx)!r}.")
        self.model_onnx = model_onnx

    @property
    def input_names(self):
        "Returns input names."
        return [i.name for i in self.model_onnx.graph.input]

    @property
    def output_names(self):
        "Returns output names."
        return [i.name for i in self.model_onnx.graph.output]

    def run(self, inputs):
        """
        Computes the outputs of the graph.

        :param inputs: dictionary
        :return: all intermediates results and output as a dictionary
        """
        if not isinstance(inputs, dict):
            raise TypeError(
                f"inputs must be a dictionary not {type(inputs)!r}.")
        results = inputs.copy()

        for init in self.model_onnx.graph.initializer:
            name = init.name
            mat = _var_as_dict(init)['value']
            results[name] = mat

        for node in self.model_onnx.graph.node:
            op_type = node.op_type
            inp = [results[n] for n in node.input]
            meth_name = f"_op_{op_type.lower()}"
            if not hasattr(self, meth_name):
                raise NotImplementedError(
                    f"OnnxMicroRuntime does not implement operator {op_type!r}.")
            kwargs = {}
            for at in node.attribute:
                var = _var_as_dict(at)
                kwargs[at.name] = var['value']
            out = getattr(self, meth_name)(*inp, **kwargs)
            for n, o in zip(node.output, out):
                results[n] = o

        return results

    ########################
    # Runtime for operators
    ########################

    def _op_abs(self, x):
        "Runtime for operator :epkg:`Op:Abs`."
        return (numpy.abs(x), )

    def _op_add(self, x, y):
        "Runtime for operator :epkg:`Op:Add`."
        return (x + y, )

    def _op_concat(self, *args, axis=None):
        "Runtime for operator :epkg:`Op:Concat`."
        def _preprocess(a, axis):
            if axis >= len(a.shape):
                new_shape = a.shape + (1, ) * (axis + 1 - len(a.shape))
                return a.reshape(new_shape)
            return a

        targs = tuple(_preprocess(a, axis) for a in args)
        return (numpy.concatenate(targs, axis), )

    def _op_gemm(self, a, b, c=None, alpha=None, beta=None,
                 transA=False, transB=False):
        "Runtime for operator :epkg:`Op:Gemm`."

        def _gemm00(a, b, c, alpha, beta):
            o = numpy.dot(a, b) * alpha
            if beta != 0:
                o += c * beta
            return o

        def _gemm01(a, b, c, alpha, beta):
            o = numpy.dot(a, b.T) * alpha
            if beta != 0:
                o += c * beta
            return o

        def _gemm10(a, b, c, alpha, beta):
            o = numpy.dot(a.T, b) * alpha
            if beta != 0:
                o += c * beta
            return o

        def _gemm11(a, b, c, alpha, beta):
            o = numpy.dot(a.T, b.T) * alpha
            if beta != 0:
                o += c * beta
            return o

        if not isinstance(transA, (int, bool, numpy.int64)):
            raise TypeError(  # pragma: no cover
                f"Unexpected type for transA: {type(transA)!r}.")
        if not isinstance(transB, (int, bool, numpy.int64)):
            raise TypeError(  # pragma: no cover
                f"Unexpected type for transA: {type(transB)!r}.")
        if transA:
            fct = _gemm11 if transB else _gemm10
        else:
            fct = _gemm01 if transB else _gemm00
        return (fct(a, b, c, alpha=alpha, beta=beta), )

    def _op_gather(self, x, indices, axis=None):
        "Runtime for operator :epkg:`Op:Gather`."
        if not x.flags['C_CONTIGUOUS']:
            x = numpy.ascontiguousarray(x)
        if not indices.flags['C_CONTIGUOUS']:
            indices = indices.ascontiguousarray()
        return (numpy.take(x, indices, axis=axis), )

    def _op_identity(self, x):
        "Runtime for operator :epkg:`Op:Identity`."
        return (x, )

    def _op_matmul(self, x, y):
        "Runtime for operator :epkg:`Op:MatMul`."
        return (numpy.matmul(x, y), )

    def _op_max(self, *inps):
        "Runtime for operator :epkg:`Op:Max`."
        return (numpy.maximum(*inps), )

    def _op_mul(self, x, y):
        "Runtime for operator :epkg:`Op:Mul`."
        return (x * y, )

    def _op_reduceprod(self, data, axes=None, keepdims=None):
        "Runtime for operator :epkg:`Op:ReduceProd`."
        if axes is not None and not isinstance(axes, int):
            if isinstance(axes, numpy.ndarray) and len(axes.shape) == 0:
                axes = int(axes)
            else:
                axes = tuple(axes) if len(axes) > 0 else None
        return (numpy.prod(data, axis=axes,
                           keepdims=keepdims,
                           dtype=data.dtype), )

    def _op_reducesum(self, data, axes, keepdims=None,
                      noop_with_empty_axes=None):
        "Runtime for operator :epkg:`Op:ReduceSum`."
        if axes is None and noop_with_empty_axes:
            return (data, )
        if axes is not None and not isinstance(axes, int):
            if isinstance(axes, numpy.ndarray) and len(axes.shape) == 0:
                axes = int(axes)
            else:
                axes = tuple(axes) if len(axes) > 0 else None
        return (numpy.sum(data, axis=axes,
                          keepdims=keepdims,
                          dtype=data.dtype), )

    def _op_reshape(self, x, shape):
        "Runtime for operator :epkg:`Op:Reshape`."
        return (x.reshape(shape), )

    def _op_shape(self, x):
        "Runtime for operator :epkg:`Op:Shape`."
        return (numpy.array(list(x.shape), dtype=numpy.int64), )

    def _op_squeeze(self, x, axes=None):
        "Runtime for operator :epkg:`Op:Squeeze`."
        if axes is None:
            return (x, )
        if hasattr(axes, '__iter__'):
            return (numpy.squeeze(x, axis=tuple(axes)), )
        return (numpy.squeeze(x, axis=axes), )

    def _op_transpose(self, x, perm=None):
        "Runtime for operator :epkg:`Op:Transpose`."
        return (numpy.transpose(x, perm), )

    def _op_unsqueeze(self, x, axes=None):
        "Runtime for operator :epkg:`Op:Unsqueeze`."
        if axes is None:
            return (x, )
        if hasattr(axes, '__iter__'):
            return (numpy.expand_dims(x, axis=tuple(axes)), )
        return (numpy.expand_dims(x, axis=axes), )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnary


class Flatten(OpRunUnary):

    atts = {'axis': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=Flatten.atts,
                            **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        i = self.axis
        shape = x.shape
        new_shape = ((1, -1) if i == 0 else
                     (numpy.prod(shape[:i]).astype(int), -1))
        return (x.reshape(new_shape), )

    def to_python(self, inputs):
        lines = ['new_shape = ((1, -1) if axis == 0 else',
                 f'    (numpy.prod({inputs[0]}.shape[:axis]).astype(int), -1))',
                 f'return {inputs[0]}.reshape(new_shape)']
        return 'import numpy', '\n'.join(lines)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from collections import OrderedDict
import numpy
from onnx.defs import onnx_opset_version
from ._op_helper import _get_typed_class_attribute
from ._op import OpRunUnaryNum, RuntimeTypeError
from ._new_ops import OperatorSchema
from .op_tree_ensemble_regressor_ import (  # pylint: disable=E0611,E0401
    RuntimeTreeEnsembleRegressorFloat, RuntimeTreeEnsembleRegressorDouble)
from .op_tree_ensemble_regressor_p_ import (  # pylint: disable=E0611,E0401
    RuntimeTreeEnsembleRegressorPFloat, RuntimeTreeEnsembleRegressorPDouble)


class TreeEnsembleRegressorCommon(OpRunUnaryNum):

    def __init__(self, dtype, onnx_node, desc=None,
                 expected_attributes=None, runtime_version=3, **options):
        OpRunUnaryNum.__init__(
            self, onnx_node, desc=desc,
            expected_attributes=expected_attributes, **options)
        self._init(dtype=dtype, version=runtime_version)

    def _get_typed_attributes(self, k):
        return _get_typed_class_attribute(self, k, self.__class__.atts)

    def _find_custom_operator_schema(self, op_name):
        """
        Finds a custom operator defined by this runtime.
        """
        if op_name == "TreeEnsembleRegressorDouble":
            return TreeEnsembleRegressorDoubleSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _init(self, dtype, version):
        atts = []
        for k in self.__class__.atts:
            v = self._get_typed_attributes(k)
            if k.endswith('_as_tensor'):
                if (v is not None and isinstance(v, numpy.ndarray) and
                        v.size > 0):
                    # replacements
                    atts[-1] = v
                    if dtype is None:
                        dtype = v.dtype
                continue
            atts.append(v)

        if dtype is None:
            dtype = numpy.float32

        if dtype == numpy.float32:
            if version == 0:
                self.rt_ = RuntimeTreeEnsembleRegressorFloat()
            elif version == 1:
                self.rt_ = RuntimeTreeEnsembleRegressorPFloat(
                    60, 20, False, False)
            elif version == 2:
                self.rt_ = RuntimeTreeEnsembleRegressorPFloat(
                    60, 20, True, False)
            elif version == 3:
                self.rt_ = RuntimeTreeEnsembleRegressorPFloat(
                    60, 20, True, True)
            else:
                raise ValueError(f"Unknown version '{version}'.")
        elif dtype == numpy.float64:
            if version == 0:
                self.rt_ = RuntimeTreeEnsembleRegressorDouble()
            elif version == 1:
                self.rt_ = RuntimeTreeEnsembleRegressorPDouble(
                    60, 20, False, False)
            elif version == 2:
                self.rt_ = RuntimeTreeEnsembleRegressorPDouble(
                    60, 20, True, False)
            elif version == 3:
                self.rt_ = RuntimeTreeEnsembleRegressorPDouble(
                    60, 20, True, True)
            else:
                raise ValueError(f"Unknown version '{version}'.")
        else:
            raise RuntimeTypeError(  # pragma: no cover
                f"Unsupported dtype={dtype}.")
        self.rt_.init(*atts)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        This is a C++ implementation coming from
        :epkg:`onnxruntime`.
        `tree_ensemble_classifier.cc
        <https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/ml/tree_ensemble_classifier.cc>`_.
        See class :class:`RuntimeTreeEnsembleRegressorFloat
        <mlprodict.onnxrt.ops_cpu.op_tree_ensemble_regressor_.RuntimeTreeEnsembleRegressorFloat>` or
        class :class:`RuntimeTreeEnsembleRegressorDouble
        <mlprodict.onnxrt.ops_cpu.op_tree_ensemble_regressor_.RuntimeTreeEnsembleRegressorDouble>`.
        """
        if hasattr(x, 'todense'):
            x = x.todense()
        pred = self.rt_.compute(x)
        if pred.shape[0] != x.shape[0]:
            pred = pred.reshape(x.shape[0], pred.shape[0] // x.shape[0])
        return (pred, )


class TreeEnsembleRegressor_1(TreeEnsembleRegressorCommon):

    atts = OrderedDict([
        ('aggregate_function', b'SUM'),
        ('base_values', numpy.empty(0, dtype=numpy.float32)),
        ('base_values_as_tensor', []),
        ('n_targets', 1),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float32)),
        ('post_transform', b'NONE'),
        ('target_ids', numpy.empty(0, dtype=numpy.int64)),
        ('target_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_weights', numpy.empty(0, dtype=numpy.float32)),
    ])

    def __init__(self, onnx_node, desc=None, runtime_version=1, **options):
        TreeEnsembleRegressorCommon.__init__(
            self, numpy.float32, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleRegressor_1.atts,
            runtime_version=runtime_version, **options)


class TreeEnsembleRegressor_3(TreeEnsembleRegressorCommon):

    atts = OrderedDict([
        ('aggregate_function', b'SUM'),
        ('base_values', numpy.empty(0, dtype=numpy.float32)),
        ('base_values_as_tensor', []),
        ('n_targets', 1),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_hitrates_as_tensor', []),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_values_as_tensor', []),
        ('post_transform', b'NONE'),
        ('target_ids', numpy.empty(0, dtype=numpy.int64)),
        ('target_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_weights', numpy.empty(0, dtype=numpy.float32)),
        ('target_weights_as_tensor', []),
    ])

    def __init__(self, onnx_node, desc=None, runtime_version=1, **options):
        TreeEnsembleRegressorCommon.__init__(
            self, None, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleRegressor_3.atts,
            runtime_version=runtime_version, **options)


class TreeEnsembleRegressorDouble(TreeEnsembleRegressorCommon):
    """
    Runtime for the custom operator `TreeEnsembleRegressorDouble`.
    .. exref::
        :title: How to use TreeEnsembleRegressorDouble instead of TreeEnsembleRegressor
        .. runpython::
            :showcode:
            import warnings
            import numpy
            from sklearn.datasets import make_regression
            from sklearn.ensemble import (
                RandomForestRegressor, GradientBoostingRegressor,
                HistGradientBoostingRegressor)
            from mlprodict.onnx_conv import to_onnx
            from mlprodict.onnxrt import OnnxInference
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                models = [
                    RandomForestRegressor(n_estimators=10),
                    GradientBoostingRegressor(n_estimators=10),
                    HistGradientBoostingRegressor(max_iter=10),
                ]
                X, y = make_regression(1000, n_features=5, n_targets=1)
                X = X.astype(numpy.float64)
                conv = {}
                for model in models:
                    model.fit(X[:500], y[:500])
                    onx64 = to_onnx(model, X, rewrite_ops=True, target_opset=15)
                    assert 'TreeEnsembleRegressorDouble' in str(onx64)
                    expected = model.predict(X)
                    oinf = OnnxInference(onx64)
                    got = oinf.run({'X': X})
                    diff = numpy.abs(got['variable'] - expected)
                    print("%s: max=%f mean=%f" % (
                        model.__class__.__name__, diff.max(), diff.mean()))
    """

    atts = OrderedDict([
        ('aggregate_function', b'SUM'),
        ('base_values', numpy.empty(0, dtype=numpy.float64)),
        ('n_targets', 1),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float64)),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float64)),
        ('post_transform', b'NONE'),
        ('target_ids', numpy.empty(0, dtype=numpy.int64)),
        ('target_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('target_weights', numpy.empty(0, dtype=numpy.float64)),
    ])

    def __init__(self, onnx_node, desc=None, runtime_version=1, **options):
        TreeEnsembleRegressorCommon.__init__(
            self, numpy.float64, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleRegressorDouble.atts,
            runtime_version=runtime_version, **options)


class TreeEnsembleRegressorDoubleSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleRegressorDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'TreeEnsembleRegressorDouble')
        self.attributes = TreeEnsembleRegressorDouble.atts


if onnx_opset_version() >= 16:
    TreeEnsembleRegressor = TreeEnsembleRegressor_3
else:
    TreeEnsembleRegressor = TreeEnsembleRegressor_1



"""
@file
@brief Inspired from sklearn-onnx, handles two backends.
"""
import numpy
from .utils_backend_onnxruntime import _capture_output


from .tests_helper import (  # noqa
    binary_array_to_string,
    dump_data_and_model,
    dump_one_class_classification,
    dump_binary_classification,
    dump_multilabel_classification,
    dump_multiple_classification,
    dump_multiple_regression,
    dump_single_regression,
    convert_model,
    fit_classification_model,
    fit_classification_model_simple,
    fit_multilabel_classification_model,
    fit_regression_model)


def create_tensor(N, C, H=None, W=None):
    "Creates a tensor."
    if H is None and W is None:
        return numpy.random.rand(N, C).astype(numpy.float32, copy=False)  # pylint: disable=E1101
    elif H is not None and W is not None:
        return numpy.random.rand(N, C, H, W).astype(numpy.float32, copy=False)  # pylint: disable=E1101
    raise ValueError(  # pragma no cover
        'This function only produce 2-D or 4-D tensor.')


def ort_version_greater(ver):
    """
    Tells if onnxruntime version is greater than *ver*.

    :param ver: version as a string
    :return: boolean
    """
    from onnxruntime import __version__
    from pyquickhelper.texthelper.version_helper import compare_module_version
    return compare_module_version(__version__, ver) >= 0



"""
@file
@brief Datasets to tests models.
"""
import os
from pandas import read_csv


def load_audit():
    """
    Use to test conversion of
    :epkg:`sklearn:ensemble:GradientBoostingClassifier`
    into :epkg:`ONNX`.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.data import load_audit
        df = load_audit()
        print(df.head())
    """
    name = os.path.dirname(__file__)
    name = os.path.join(name, 'audit.csv')
    df = read_csv(name).drop(['ID', 'index'], axis=1, inplace=False).dropna()
    return df



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.

.. versionadded:: 0.9
"""
from ._op import OpRun


class SequenceEmpty(OpRun):

    atts = {}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       atts=SequenceEmpty.atts, **options)

    def _run(self, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return ([], )



"""
@file
@brief Command line about exporting :epkg:`asv` results into a dataframe.
"""
from datetime import datetime
from ..asv_benchmark.asv_exports import enumerate_export_asv_json


def asv2csv(folder, outfile=None, last_one=False, baseline=None,
            conf=None, fLOG=print):
    """
    Converts results produced by :epkg:`asv` into :epkg:`csv`.

    :param folder: folder where the results are
    :param outfile: output the results into :epkg:`csv`
    :param last_one: converts only the last report into csv
    :param baseline: baseline usually ``'skl'``, if not empty,
        computes ratios
    :param conf: test configuration, to retrieve more metadata
    :param fLOG: logging function

    .. cmdref::
        :title: Converts asv results into csv
        :cmd: -m mlprodict asv2csv--help
        :lid: l-cmd-asv2csv

        The command converts :epkg:`asv` results into :epkg:`csv`.

        Example::

            python -m mlprodict asv2csv -f <folder> -o result.csv

    The filename may contain ``<date>``, it is then replaced by
    the time now.
    """
    iter_rows = enumerate_export_asv_json(
        folder, last_one=last_one, baseline=baseline, conf=conf)

    if outfile is None:
        rows = []
        for row in iter_rows:
            fLOG(row)
            rows.append(row)
        return rows

    import pandas
    df = pandas.DataFrame(iter_rows)
    outfile = outfile.replace(
        "<date>",
        datetime.now().strftime("%Y%m%dT%H%M%S"))
    df.to_csv(outfile, index=False)
    return df



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *sklapi*.
Importing this file imports :epkg:`sklearn-onnx` as well.
"""
from .onnx_pipeline import OnnxPipeline
from .onnx_transformer import OnnxTransformer
from .onnx_speed_up import (
    OnnxSpeedupClassifier,
    OnnxSpeedupCluster,
    OnnxSpeedupRegressor,
    OnnxSpeedupTransformer)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from scipy.spatial.distance import cdist
from ._op import OpRunBinaryNum
from ._new_ops import OperatorSchema


class CDist(OpRunBinaryNum):

    atts = {'metric': 'sqeuclidean', 'p': 2.}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNum.__init__(self, onnx_node, desc=desc,
                                expected_attributes=CDist.atts,
                                **options)

    def _run(self, a, b, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        metric = self.metric.decode('ascii')
        if metric == 'minkowski':
            res = cdist(a, b, metric=metric, p=self.p)
        else:
            res = cdist(a, b, metric=metric)
        # scipy may change the output type
        res = res.astype(a.dtype)
        return (res, )

    def _find_custom_operator_schema(self, op_name):
        if op_name == "CDist":
            return CDistSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def to_python(self, inputs):
        metric = self.metric.decode('ascii')
        if metric == 'minkowski':
            return ('from scipy.spatial.distance import cdist',
                    "return cdist({}, {}, metric='{}', p={})".format(
                        inputs[0], inputs[1], metric, self.p))
        return ('from scipy.spatial.distance import cdist',
                f"return cdist({inputs[0]}, {inputs[1]}, metric='{metric}')")


class CDistSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleClassifierDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'CDist')
        self.attributes = CDist.atts



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def scatter_elements(data, indices, updates, axis=0):
    """
    ::
        // for 3-dim and axis=0
        //    output[indices[i][j][k]][j][k] = updates[i][j][k]
        // for axis 1
        //    output[i][indices[i][j][k]][k] = updates[i][j][k]
        // and so on
    """
    if len(data.shape) == 1 and axis == 0:
        scattered = numpy.copy(data)
        for pos, up in zip(indices, updates):
            scattered[pos] = up
        return scattered

    if axis < 0:
        axis = data.ndim + axis

    idx_xsection_shape = indices.shape[:axis] + indices.shape[axis + 1:]

    def make_slice(arr, axis, i):
        slc = [slice(None)] * arr.ndim
        slc[axis] = i
        return slc

    def unpack(packed):
        unpacked = packed[0]
        for i in range(1, len(packed)):
            unpacked = unpacked, packed[i]
        return unpacked

    # We use indices and axis parameters to create idx
    # idx is in a form that can be used as a NumPy advanced
    # indices for scattering of updates param. in data
    idx = [[unpack(numpy.indices(idx_xsection_shape).reshape(indices.ndim - 1, -1)),
            indices[tuple(make_slice(indices, axis, i))].reshape(1, -1)[0]]
           for i in range(indices.shape[axis])]
    idx = list(numpy.concatenate(idx, axis=1))
    idx.insert(axis, idx.pop())

    # updates_idx is a NumPy advanced indices for indexing
    # of elements in the updates
    updates_idx = list(idx)
    updates_idx.pop(axis)
    updates_idx.insert(axis, numpy.repeat(numpy.arange(indices.shape[axis]),
                                          numpy.prod(idx_xsection_shape)))

    scattered = numpy.copy(data)
    scattered[tuple(idx)] = updates[tuple(updates_idx)]
    return scattered


class ScatterElements(OpRun):

    atts = {'axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, data, indices, updates, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        res = scatter_elements(data, indices, updates, axis=self.axis)
        return (res, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from .op_non_max_suppression_ import RuntimeNonMaxSuppression  # pylint: disable=E0611


class NonMaxSuppression(OpRun):

    atts = {'center_point_box': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=NonMaxSuppression.atts,
                       **options)
        self.inst = RuntimeNonMaxSuppression()
        self.inst.init(self.center_point_box)

    def _run(self, boxes, scores, max_output_boxes_per_class=None,  # pylint: disable=W0221
             iou_threshold=None, score_threshold=None,
             attributes=None, verbose=0, fLOG=None):
        if max_output_boxes_per_class is None:
            max_output_boxes_per_class = numpy.array([], dtype=numpy.int64)
        if iou_threshold is None:
            iou_threshold = numpy.array([], dtype=numpy.float32)
        if score_threshold is None:
            score_threshold = numpy.array([], dtype=numpy.float32)
        res = self.inst.compute(boxes, scores, max_output_boxes_per_class,
                                iou_threshold, score_threshold)
        res = res.reshape((-1, 3))
        return (res, )



"""
@file
@brief Computes shape inference for onnx operators.
"""
from .shape_excs import ShapeInferenceException, ShapeInferenceDimensionError
from .shape_result import (
    ShapeResult, OnnxKind, ShapeConstraintList, ShapeConstraint)


def shape_det(known_shapes, node):
    "Infers shape for operator Abs."
    x = known_shapes[node.input[0]]
    if x.mtype != OnnxKind.Tensor:
        raise ShapeInferenceException(  # pragma: no cover
            f"Result {x!r} must be a tensor.")
    if x.n_dims() < 2:
        if x.n_dims() > 0:
            raise ShapeInferenceException(  # pragma: no cover
                f"Operator Det requires at least two dimensions not {x.n_dims()!r}.")
        raise ShapeInferenceDimensionError(  # pragma: no cover
            f"Operator Det requires at least two dimensions not {x.n_dims()!r}.")
    name = node.output[0]

    constraints = ShapeConstraintList()
    a, b = x.shape[-2:]
    if isinstance(a, int) and isinstance(b, int):
        if a != b:
            raise ShapeInferenceException(  # pragma: no cover
                f"Operator Det only applies on square matrices not {x.n_dims()!r}.")
    elif isinstance(a, str):
        constraints.append(ShapeConstraint(a, {b}))
    elif isinstance(b, str):
        constraints.append(ShapeConstraint(b, {a}))
    else:
        raise ShapeInferenceException(  # pragma: no cover
            f"Unexpected case for operator Det ({x!r}).")
    if x.n_dims() == 2:
        r = ShapeResult(name, [], x.dtype, False,
                        x.mtype, constraints)
    else:
        r = ShapeResult(name, x.shape[:-2], x.dtype, False,
                        x.mtype, constraints)
    return known_shapes.update(name, r)



"""
@file
@brief Shortcut to *ops_shape*.
"""
import textwrap
from onnx.onnx_cpp2py_export.defs import SchemaError  # pylint: disable=E0401,E0611
from ...onnx_tools.onnx2py_helper import get_onnx_schema
from .shape_excs import ShapeInferenceMissing
from ._element_unary import (
    shape_abs, shape_acos, shape_acosh,
    shape_asin, shape_asinh, shape_atan, shape_atanh,
    shape_castlike, shape_ceil, shape_celu,
    shape_clip, shape_cos, shape_cosh,
    shape_elu, shape_erf, shape_exp, shape_floor,
    shape_hardmax, shape_hardsigmoid,
    shape_identity, shape_isinf, shape_isnan,
    shape_leakyrelu, shape_log, shape_logsoftmax,
    shape_neg, shape_not, shape_reciprocal, shape_relu, shape_round,
    shape_selu, shape_shrink,
    shape_sigmoid, shape_sign, shape_sin, shape_sinh, shape_softmax,
    shape_softplus, shape_softsign,
    shape_sqrt, shape_tan, shape_tanh, shape_thresholdedrelu,
    shape_trilu)
from ._element_wise import (
    shape_add, shape_and,
    shape_div,
    shape_equal,
    shape_greater, shape_greaterorequal,
    shape_less, shape_lessorequal,
    shape_max, shape_min, shape_mod, shape_mul,
    shape_or,
    shape_pow,
    shape_sub,
    shape_xor)
from ._op_shape_op import shape_det


_shape_functions = {
    k: v for k, v in globals().items() if k.startswith("shape_")
}


count = [0]


def shape_dispatch(cache, known_shape, node, rt_class=None):
    """
    Calls the corresponding fucntion for every node.

    :param cache: cache used function
    :param known_shape: known_shape for all results
    :param node: onnx node
    :param rt_class: a node may be a predefined function in onnx,
        if no specific function is available, the predefined
        onnx definition is used and run through this runtime
    :return: was *known_shape* updated or not...
    """
    key = node.domain, node.op_type
    fct_shape = None
    if key in cache:
        fct_shape = cache[key]
    else:
        op_type = "shape_" + node.op_type.lower()
        if op_type in _shape_functions:
            fct_shape = _shape_functions[op_type]
            cache[key] = fct_shape

    if fct_shape is None and rt_class is not None:
        # check this operator is a predefined function in ONNX.
        try:
            onnx_schema = get_onnx_schema(node.op_type, node.domain)
        except SchemaError:
            onnx_schema = None
        if onnx_schema is not None and onnx_schema.has_function:
            sess = rt_class(onnx_schema.function_body)
            if len(node.input) != len(sess.input_names):
                raise RuntimeError(  # pragma: no cover
                    "node and function must have the same number of inputs, "
                    "len(%r) != len(%r)." % (
                        node.input, sess.input_names))
            if len(node.output) != len(sess.output_names):
                raise RuntimeError(  # pragma: no cover
                    "node and function must have the same number of outputs, "
                    "len(%r) != len(%r)." % (
                        node.output, sess.output_names))

            def _shape_function(known_shape, node):
                inputs = {iname: known_shape[name] for name, iname in
                          zip(node.input, sess.input_names)}
                outputs = sess.run(inputs)
                res = False
                for name, oname in zip(node.output, sess.output_names):
                    r = known_shape.update(name, outputs[oname])
                    res = res or r
                return res

            fct_shape = _shape_function
            cache[key] = fct_shape

    if fct_shape is not None:
        return fct_shape(known_shape, node)

    raise ShapeInferenceMissing(  # pragma: no cover
        "Unable to find a corresponding function for operator type %r "
        "domain=%r, looking for %r among\n%s" % (
            node.op_type, node.domain, "shape_" + node.op_type.lower(),
            "\n".join(textwrap.wrap(
                " ".join(_ for _ in sorted(_shape_functions))))))
